import copy
import functools
import os
import threading
import urllib
import uuid

import numpy as np
from astropy import units as u
from astropy.coordinates import SpectralCoord
from astropy.io import fits
from astropy.utils.data import download_file
from astropy.wcs import WCS
from astropy_healpix import HEALPix, level_to_nside
from dask import array as da

from ._high_level import VALID_COORD_SYSTEM
from ._trim_utils import fits_getdata_untrimmed
from ._utils import (
    is_url,
    load_properties,
    map_header,
    skycoord_first,
    spectral_coord_to_index,
    tile_filename,
)

__all__ = ["hips_as_dask_array"]


class HiPSArray:

    def __init__(self, directory_or_url, level=None, level_depth=None):

        # We strip any trailing slashes since we then assume in the rest of the
        # code that we need to add a slash (and double slashes cause issues for URLs)
        self._directory_or_url = str(directory_or_url).rstrip("/")

        self._is_url = is_url(self._directory_or_url)

        self._properties = load_properties(self._directory_or_url)

        if self._properties["dataproduct_type"] == "image":
            self.ndim = 2
        elif self._properties["dataproduct_type"] == "spectral-cube":
            self.ndim = 3
        else:
            raise TypeError(f"HiPS type {self._properties['dataproduct_type']} not recognized")

        self._tile_width = int(self._properties["hips_tile_width"])
        self._order_spatial = int(self._properties["hips_order"])

        if level is None:
            self._level_spatial = self._order_spatial
        else:
            if level > self._order_spatial:
                raise ValueError(
                    f"HiPS dataset at {self._directory_or_url} does not contain spatial level {level} data"
                )
            elif level < 0:
                raise ValueError("level should be positive")
            else:
                self._level_spatial = int(level)

        if self.ndim == 3:

            # TODO: here need to check consistency, maybe actually don't allow spectral level to be passed in

            self._tile_depth = int(self._properties["hips_tile_depth"])
            try:
                self._order_depth = int(self._properties["hips_order_axis2"])
            except KeyError:
                # This property was originally called hips_order_freq, so try that for backward-compatibility
                self._order_depth = int(self._properties["hips_order_freq"])

            if level_depth is None:
                self._level_depth = self._order_depth - (self._order_spatial - self._level_spatial)
            else:
                if level_depth > self._order_depth:
                    raise ValueError(
                        f"HiPS dataset at {self._directory_or_url} does not contain spectral level {level_depth} data"
                    )
                elif level_depth < 0:
                    raise ValueError("level_depth should be positive")
                else:
                    self._level_depth = int(level_depth)

            self._level = (self._level_spatial, self._level_depth)
            self._tile_dims = (self._tile_width, self._tile_depth)

        else:

            self._level_depth = None
            self._level = self._level_spatial
            self._tile_dims = self._tile_width

        self._tile_format = self._properties["hips_tile_format"]
        self._frame_str = self._properties["hips_frame"]
        self._frame = VALID_COORD_SYSTEM[self._frame_str]

        self._hp = HEALPix(
            nside=level_to_nside(self._level_spatial), frame=self._frame, order="nested"
        )

        self._header = map_header(level=self._level, frame=self._frame, tile_dims=self._tile_dims)

        self.wcs = WCS(self._header)
        self.shape = self.wcs.array_shape

        # Determine actual spectral range, because we don't actually want to
        # create a dask array with the full possible range of spectral indices
        # since this will be huge and unnecessary

        if self.ndim == 3:

            wav_min = SpectralCoord(float(self._properties["em_min"]), u.m)
            wav_max = SpectralCoord(float(self._properties["em_max"]), u.m)

            index_min = spectral_coord_to_index(self._level_depth, wav_min)
            index_max = spectral_coord_to_index(self._level_depth, wav_max)

            if index_min > index_max:
                index_min, index_max = index_max, index_min

            index_max += 1

            index_min *= self._tile_depth
            index_max *= self._tile_depth

            self.wcs = self.wcs[index_min:index_max]
            self.shape = (index_max - index_min,) + self.shape[1:]

        # FIX following
        self.dtype = float

        if self.ndim == 2:
            self.chunksize = (self._tile_width, self._tile_width)
        else:
            self.chunksize = (self._tile_depth, self._tile_width, self._tile_width)

        self._nan = np.nan * np.ones(self.chunksize, dtype=self.dtype)

        self._blank = np.broadcast_to(np.nan, self.shape)

        # astropy.wcs (wcslib) is not thread-safe, so when dask computes chunks
        # in parallel each thread needs to use its own copy of the WCS.
        self._local = threading.local()

    @property
    def _thread_wcs(self):
        # Return a copy of the WCS that is unique to the calling thread.
        wcs = getattr(self._local, "wcs", None)
        if wcs is None:
            if hasattr(self.wcs, "deepcopy"):
                wcs = self.wcs.deepcopy()
            else:
                wcs = copy.deepcopy(self.wcs)
            self._local.wcs = wcs
        return wcs

    def __getitem__(self, item):

        for idx in range(self.ndim):
            if item[idx].start == item[idx].stop:
                return self._blank[item]

        # Determine spatial healpix index - we use two points in different
        # parts of the image because in some cases using the exact center or
        # corners can cause issues.

        istart = item[-2].start
        irange = item[-2].stop - item[-2].start
        imid = np.array([istart + 0.25 * irange, istart + 0.75 * irange])

        jstart = item[-1].start
        jrange = item[-1].stop - item[-1].start
        jmid = np.array([jstart + 0.25 * jrange, jstart + 0.75 * jrange])

        # Convert pixel coordinates to HEALPix indices

        # astropy.wcs is not thread-safe, so use this thread's own WCS copy
        wcs = self._thread_wcs

        if self.ndim == 2:
            coord = wcs.pixel_to_world(jmid, imid)
        else:
            kmid = 0.5 * (item[0].start + item[0].stop)
            coord, spectral_coord = skycoord_first(wcs.pixel_to_world(jmid, imid, kmid))

        if self._frame_str == "equatorial":
            lon, lat = coord.ra.deg, coord.dec.deg
        elif self._frame_str == "galactic":
            lon, lat = coord.l.deg, coord.b.deg
        else:
            raise NotImplementedError()

        invalid = np.isnan(lon) | np.isnan(lat)

        if np.all(invalid):
            return self._nan
        elif np.any(invalid):
            coord = coord[~invalid]

        spatial_index = self._hp.skycoord_to_healpix(coord)

        if np.all(spatial_index == -1):
            return self._nan

        spatial_index = np.max(spatial_index)

        # Determine spectral index, if needed
        if self.ndim == 3:
            spectral_index = spectral_coord_to_index(self._level_depth, spectral_coord).max()
            index = (spatial_index, spectral_index)
        else:
            index = spatial_index

        return self._get_tile(level=self._level, index=index).astype(float)

    @functools.lru_cache(maxsize=128)  # noqa: B019
    def _get_tile(self, *, level, index):

        filename_or_url = tile_filename(
            level=self._level,
            index=index,
            output_directory=self._directory_or_url,
            extension="fits",
        )

        if self._is_url:
            try:
                filename = download_file(filename_or_url, cache=True)
            except urllib.error.HTTPError:
                return self._nan
        elif not os.path.exists(filename_or_url):
            return self._nan
        else:
            filename = filename_or_url

        if self.ndim == 2:
            return fits.getdata(filename)
        else:
            return fits_getdata_untrimmed(
                filename,
                tile_size=self._tile_width,
                tile_depth=self._tile_depth,
            )


def hips_as_dask_array(directory_or_url, *, level=None):
    """
    Return a dask array and WCS that represent a HiPS dataset at a particular level.
    """
    array_wrapper = HiPSArray(directory_or_url, level=level)
    return (
        da.from_array(
            array_wrapper,
            chunks=array_wrapper.chunksize,
            name=str(uuid.uuid4()),
            meta=np.array([], dtype=float),
        ),
        array_wrapper.wcs,
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import operator

import numpy as np
import pytest
from numpy.testing import assert_equal

from .._subset_array import ReprojectedArraySubset


class TestReprojectedArraySubset:
    def setup_method(self, method):
        self.array1 = np.random.random((123, 87))
        self.array2 = np.random.random((123, 87))
        self.array3 = np.random.random((123, 87))
        self.array4 = np.random.random((123, 87, 16))

        self.footprint1 = (self.array1 > 0.5).astype(int)
        self.footprint2 = (self.array2 > 0.5).astype(int)
        self.footprint3 = (self.array3 > 0.5).astype(int)
        self.footprint4 = (self.array4 > 0.5).astype(int)

        self.subset1 = ReprojectedArraySubset(
            self.array1[20:88, 34:40],
            self.footprint1[20:88, 34:40],
            [(20, 88), (34, 40)],
        )

        self.subset2 = ReprojectedArraySubset(
            self.array2[50:123, 37:42],
            self.footprint2[50:123, 37:42],
            [(50, 123), (37, 42)],
        )

        self.subset3 = ReprojectedArraySubset(
            self.array3[40:50, 11:19],
            self.footprint3[40:50, 11:19],
            [(40, 50), (11, 19)],
        )

        self.subset4 = ReprojectedArraySubset(
            self.array4[30:35, 40:45, 1:4],
            self.footprint4[30:35, 40:45, 1:4],
            [(30, 35), (40, 45), (1, 4)],
        )

    def test_repr(self):
        assert repr(self.subset1) == "<ReprojectedArraySubset at [20:88,34:40]>"

    def test_view_in_original_array(self):
        assert_equal(self.array1[self.subset1.view_in_original_array], self.subset1.array)
        assert_equal(self.footprint1[self.subset1.view_in_original_array], self.subset1.footprint)

    def test_shape(self):
        assert self.subset1.shape == (68, 6)

    def test_overlaps(self):
        assert self.subset1.overlaps(self.subset1)
        assert self.subset1.overlaps(self.subset2)
        assert not self.subset1.overlaps(self.subset3)
        assert self.subset2.overlaps(self.subset1)
        assert self.subset2.overlaps(self.subset2)
        assert not self.subset2.overlaps(self.subset3)
        assert not self.subset3.overlaps(self.subset1)
        assert not self.subset3.overlaps(self.subset2)
        assert self.subset3.overlaps(self.subset3)

    @pytest.mark.parametrize("op", [operator.add, operator.sub, operator.mul, operator.truediv])
    def test_arithmetic(self, op):
        subset = op(self.subset1, self.subset2)
        assert subset.bounds == [(50, 88), (37, 40)]
        expected = op(self.array1[50:88, 37:40], self.array2[50:88, 37:40])
        assert_equal(subset.array, expected)

    def test_arithmetic_nooverlap(self):
        subset = self.subset1 - self.subset3
        assert subset.bounds == [(40, 50), (34, 34)]
        assert subset.shape == (10, 0)

    def test_overlaps_dimension_mismatch(self):
        with pytest.raises(
            ValueError, match=("Mismatch in number of dimensions, expected 2 dimensions and got 3")
        ):
            self.subset1.overlaps(self.subset4)

    def test_arithmetic_dimension_mismatch(self):
        with pytest.raises(
            ValueError, match=("Mismatch in number of dimensions, expected 2 dimensions and got 3")
        ):
            self.subset1 - self.subset4



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

import os
from pathlib import Path

import dask.array as da
import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDData
from astropy.wcs import WCS
from astropy.wcs.wcsapi import HighLevelWCSMixin, SlicedLowLevelWCS
from pyavm import AVM

from .utils import as_transparent_rgb

# Backward-compatibility shim modules emit AstropyDeprecationWarning on
# import, which causes collection errors under "filterwarnings = error".
# Exclude them from pytest doctest collection.
_pkg = Path(__file__).parent
collect_ignore = [
    str(_pkg / "array_utils.py"),
    str(_pkg / "common.py"),
    str(_pkg / "wcs_utils.py"),
    str(_pkg / "adaptive" / "core.py"),
    str(_pkg / "adaptive" / "high_level.py"),
    str(_pkg / "interpolation" / "core.py"),
    str(_pkg / "interpolation" / "high_level.py"),
    str(_pkg / "spherical_intersect" / "core.py"),
    str(_pkg / "spherical_intersect" / "high_level.py"),
    str(_pkg / "spherical_intersect" / "overlap.py"),
    str(_pkg / "healpix" / "core.py"),
    str(_pkg / "healpix" / "high_level.py"),
    str(_pkg / "healpix" / "utils.py"),
    str(_pkg / "hips" / "high_level.py"),
    str(_pkg / "hips" / "utils.py"),
    str(_pkg / "mosaicking" / "coadd.py"),
    str(_pkg / "mosaicking" / "wcs_helpers.py"),
    str(_pkg / "mosaicking" / "background.py"),
    str(_pkg / "mosaicking" / "subset_array.py"),
]

try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False

os.environ["MPLBACKEND"] = "Agg"


def pytest_configure(config):
    if ASTROPY_HEADER:
        config.option.astropy_header = True

        PYTEST_HEADER_MODULES.pop("Pandas", None)
        PYTEST_HEADER_MODULES.pop("h5py", None)
        PYTEST_HEADER_MODULES.pop("Matplotlib", None)
        PYTEST_HEADER_MODULES["Astropy"] = "astropy"
        PYTEST_HEADER_MODULES["astropy-healpix"] = "astropy_healpix"
        PYTEST_HEADER_MODULES["Cython"] = "cython"

        from reproject import __version__

        TESTED_VERSIONS["reproject"] = __version__


class TestLowLevelWCS(SlicedLowLevelWCS):
    # The simplest way to get a 'pure' low level WCS is to call SlicedLowLevelWCS
    # with an ellipsis slice!

    def __init__(self, low_level_wcs):
        self._low_level_wcs = low_level_wcs
        super().__init__(low_level_wcs, Ellipsis)


class TestHighLevelWCS(HighLevelWCSMixin):
    def __init__(self, low_level_wcs):
        self._low_level_wcs = low_level_wcs

    @property
    def low_level_wcs(self):
        return self._low_level_wcs

    # FIXME: due to a bug in astropy we need world_n_dim to be defined here

    @property
    def world_n_dim(self):
        return self.low_level_wcs.world_n_dim

    @property
    def pixel_n_dim(self):
        return self.low_level_wcs.pixel_n_dim


@pytest.fixture
def simple_celestial_fits_wcs():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "RA---TAN", "DEC--TAN"
    wcs.wcs.crpix = (1, 2)
    wcs.wcs.crval = (30, 40)
    wcs.wcs.cdelt = (-0.05, 0.04)
    wcs.wcs.equinox = 2000.0
    return wcs


@pytest.fixture(params=["fits_wcs", "ape14_low_level_wcs", "ape14_high_level_wcs"])
def simple_celestial_wcs(request, simple_celestial_fits_wcs):
    if request.param == "fits_wcs":
        return simple_celestial_fits_wcs
    elif request.param == "ape14_low_level_wcs":
        return TestLowLevelWCS(simple_celestial_fits_wcs)
    elif request.param == "ape14_high_level_wcs":
        return TestHighLevelWCS(simple_celestial_fits_wcs)


def set_wcs_array_shape(wcs, shape):
    if isinstance(wcs, WCS):
        wcs._naxis = list(shape[::-1])
    elif isinstance(wcs, TestLowLevelWCS):
        wcs._low_level_wcs._naxis = list(shape[::-1])
    elif isinstance(wcs, TestHighLevelWCS):
        wcs.low_level_wcs._naxis = list(shape[::-1])


def valid_celestial_input(tmp_path, request, wcs):
    array = np.ones((30, 40))

    kwargs = {}

    # AVM only supports FITS-WCS
    if request.param in ["png", "transparent_png", "jpeg"]:
        if not isinstance(wcs, WCS):
            pytest.skip()

    if "hdu" in request.param or request.param in ["filename", "path", "path_compressed"]:
        if not isinstance(wcs, WCS):
            pytest.skip()

        hdulist = fits.HDUList(
            [
                fits.PrimaryHDU(array, wcs.to_header()),
                fits.ImageHDU(array, wcs.to_header()),
                fits.CompImageHDU(array, wcs.to_header()),
            ]
        )

    if request.param in ["filename", "path", "path_compressed"]:
        input_value = tmp_path / "test.fits"
        if request.param == "filename":
            input_value = str(input_value)
        hdulist.writeto(input_value)
        kwargs["hdu_in"] = 2 if request.param == "path_compressed" else 0
    elif request.param == "hdulist":
        input_value = hdulist
        kwargs["hdu_in"] = 1
    elif request.param == "primary_hdu":
        input_value = hdulist[0]
    elif request.param == "image_hdu":
        input_value = hdulist[1]
    elif request.param == "comp_image_hdu":
        input_value = hdulist[2]
    elif request.param == "shape_wcs_tuple":
        input_value = (array.shape, wcs)
    elif request.param == "data_wcs_tuple":
        input_value = (array, wcs)
    elif request.param == "dask_wcs_tuple":
        input_value = (da.from_array(array), wcs)
    elif request.param == "nddata":
        input_value = NDData(data=array, wcs=wcs)
    elif request.param == "nddata_dask":
        input_value = NDData(data=da.from_array(array), wcs=wcs)
    elif request.param == "wcs":
        set_wcs_array_shape(wcs, array.shape)
        input_value = wcs
    elif request.param == "shape_wcs_tuple":
        input_value = (array.shape, wcs)
    elif request.param in ["png", "transparent_png", "jpeg"]:
        extension = "jpg" if request.param == "jpeg" else "png"
        channels = 4 if request.param == "transparent_png" else 3
        avm = AVM.from_wcs(wcs, shape=array.shape)
        array = np.broadcast_to(array[None], (channels,) + array.shape).copy()
        if channels == 4:
            array[3] = 1
        image = as_transparent_rgb(array)
        if "transparent" not in request.param:
            image = image.convert("RGB")
        original_rgb = tmp_path / f"original.{extension}"
        image.save(original_rgb)
        input_value = tmp_path / f"tagged.{extension}"
        avm.embed(original_rgb, input_value)
    else:
        raise ValueError(f"Unknown mode: {request.param}")

    return array, wcs, input_value, kwargs


COMMON_PARAMS = [
    "filename",
    "path",
    "path_compressed",
    "hdulist",
    "primary_hdu",
    "image_hdu",
    "comp_image_hdu",
    "data_wcs_tuple",
    "dask_wcs_tuple",
    "nddata",
    "nddata_dask",
    "png",
    "transparent_png",
    "jpeg",
]


@pytest.fixture(params=COMMON_PARAMS)
def valid_celestial_input_data(tmp_path, request, simple_celestial_wcs):
    return valid_celestial_input(tmp_path, request, simple_celestial_wcs)


@pytest.fixture(
    params=COMMON_PARAMS
    + [
        "wcs",
        "shape_wcs_tuple",
    ]
)
def valid_celestial_input_shapes(tmp_path, request, simple_celestial_wcs):
    return valid_celestial_input(tmp_path, request, simple_celestial_wcs)


@pytest.fixture(
    params=[
        "wcs_shape",
        "header",
        "header_shape",
        "wcs",
    ]
)
def valid_celestial_output_projections(request, simple_celestial_wcs):
    shape = (30, 40)
    wcs = simple_celestial_wcs

    # Rotate the WCS in case this is used for actual reprojection tests

    # wcs.wcs.pc = np.array([[np.cos(0.4), -np.sin(0.4)], [np.sin(0.4), np.cos(0.4)]])

    kwargs = {}

    if request.param == "wcs_shape":
        output_value = wcs
        kwargs["shape_out"] = shape
    elif request.param == "header":
        if not isinstance(wcs, WCS):
            pytest.skip()
        header = wcs.to_header()
        header["NAXIS"] = 2
        header["NAXIS1"] = 40
        header["NAXIS2"] = 30
        output_value = header
    elif request.param == "header_shape":
        if not isinstance(wcs, WCS):
            pytest.skip()
        output_value = wcs.to_header()
        kwargs["shape_out"] = shape
    elif request.param == "wcs":
        set_wcs_array_shape(wcs, (30, 40))
        output_value = wcs

    return wcs, shape, output_value, kwargs


DATA = os.path.join(os.path.dirname(__file__), "tests", "data")


@pytest.fixture(params=["fits", "asdf"])
def aia_test_data(request):

    pytest.importorskip("sunpy", minversion="6.0.1")

    from sunpy.coordinates.ephemeris import get_body_heliographic_stonyhurst
    from sunpy.map import Map

    if request.param == "fits":
        map_aia = Map(os.path.join(DATA, "aia_171_level1.fits"))
        data = map_aia.data
        wcs = map_aia.wcs
        date = map_aia.reference_date
        target_wcs = wcs.deepcopy()
    elif request.param == "asdf":
        pytest.importorskip("astropy", minversion="4.0")
        pytest.importorskip("gwcs", minversion="0.12")
        asdf = pytest.importorskip("asdf")
        aia = asdf.open(os.path.join(DATA, "aia_171_level1.asdf"))
        data = aia["data"][...]
        wcs = aia["wcs"]
        date = wcs.output_frame.reference_frame.obstime
        target_wcs = Map(os.path.join(DATA, "aia_171_level1.fits")).wcs.deepcopy()
    else:
        raise ValueError("file_format should be fits or asdf")

    # Reproject to an observer on Venus

    target_wcs.wcs.cdelt = ([24, 24] * u.arcsec).to(u.deg)
    target_wcs.wcs.crpix = [64, 64]
    venus = get_body_heliographic_stonyhurst("venus", date)
    target_wcs.wcs.aux.hgln_obs = venus.lon.to_value(u.deg)
    target_wcs.wcs.aux.hglt_obs = venus.lat.to_value(u.deg)
    target_wcs.wcs.aux.dsun_obs = venus.radius.to_value(u.m)

    return data, wcs, target_wcs



import numpy as np
from dask_image.ndinterp import map_coordinates as dask_image_map_coordinates
from dask_image.ndinterp import spline_filter
from scipy.ndimage import spline_filter as scipy_spline_filter

__all__ = ["map_coordinates", "dask_map_coordinates", "sample_array_edges", "ArrayWrapper"]


def find_chunk_shape(shape, max_chunk_size=None):
    """
    Given the shape of an n-dimensional array, and the maximum number of
    elements in a chunk, return the largest chunk shape to use for iteration.

    This currently assumes the optimal chunk shape to return is for C-contiguous
    arrays.

    Parameters
    ----------
    shape : iterable
        The shape of the n-dimensional array.
    max_chunk_size : int, optional
        The maximum number of elements per chunk.
    """

    if max_chunk_size is None:
        return tuple(shape)

    block_shape = []

    max_repeat_remaining = max_chunk_size

    for size in shape[::-1]:
        if max_repeat_remaining > size:
            block_shape.append(size)
            max_repeat_remaining = max_repeat_remaining // size
        else:
            block_shape.append(max_repeat_remaining)
            max_repeat_remaining = 1

    return tuple(block_shape[::-1])


def iterate_chunks(shape, *, max_chunk_size):
    """
    Given a data shape and a chunk shape (or maximum chunk size), iteratively
    return slice objects that can be used to slice the array.

    Parameters
    ----------
    shape : iterable
        The shape of the n-dimensional array.
    max_chunk_size : int
        The maximum number of elements per chunk.
    """

    if np.prod(shape) == 0:
        return

    chunk_shape = find_chunk_shape(shape, max_chunk_size)

    ndim = len(chunk_shape)
    start_index = [0] * ndim

    shape = list(shape)

    while start_index <= shape:
        end_index = [min(start_index[i] + chunk_shape[i], shape[i]) for i in range(ndim)]

        slices = tuple([slice(start_index[i], end_index[i]) for i in range(ndim)])

        yield slices

        # Update chunk index. What we do is to increment the
        # counter for the first dimension, and then if it
        # exceeds the number of elements in that direction,
        # cycle back to zero and advance in the next dimension,
        # and so on.
        start_index[0] += chunk_shape[0]
        for i in range(ndim - 1):
            if start_index[i] >= shape[i]:
                start_index[i] = 0
                start_index[i + 1] += chunk_shape[i + 1]

        # We can now check whether the iteration is finished
        if start_index[-1] >= shape[-1]:
            break


def at_least_float32(array):
    if array.dtype.kind == "f" and array.dtype.itemsize >= 4:
        return array
    else:
        return array.astype(np.float32)


def memory_efficient_access(array, chunk):
    # If we access a number of chunks from a memory-mapped array, memory usage
    # will increase and could crash e.g. dask.distributed workers. We therefore
    # use a temporary memmap to load the data.
    if isinstance(array, np.memmap) and array.flags.c_contiguous:
        array_tmp = np.memmap(
            array.filename,
            mode="r",
            dtype=array.dtype,
            shape=array.shape,
            offset=array.offset,
        )
        return array_tmp[chunk]
    else:
        return array[chunk]


def _clip_coords(image, coords):

    shape = image.shape

    coords = coords.copy()
    for i in range(coords.shape[0]):
        coords[i][(coords[i] < 0) & (coords[i] >= -0.5)] = 0
        coords[i][(coords[i] < shape[i] - 0.5) & (coords[i] >= shape[i] - 1)] = shape[i] - 1

    return coords


def dask_map_coordinates(image, coords, output=None, **kwargs):

    cval = kwargs.get("cval", 0.0)

    original_shape = image.shape

    # Thin wrapper around dask-image's map_coordinates which ensures that we can
    # interpolate right to the edge of the image, and also implement the output
    # keyword argument

    coords = _clip_coords(image, coords)

    if output is None:
        output = np.ones(coords.shape[1]) * cval
    else:
        output[:] = cval

    # At the time of writing, dask-image is not able to correctly handle
    # prefiltering, instead doing it per-chunk which can give subtly different
    # results
    if kwargs["order"] >= 2:
        try:
            image = spline_filter(image, order=kwargs["order"], mode="constant")
        except ValueError as exc:
            # If arrays are too small, spline_filter can fail, so we catch this
            # case and call the scipy version if so
            if "The overlapping depth" in str(exc):
                image = scipy_spline_filter(image, order=kwargs["order"], mode="constant")
            else:
                raise exc

    # dask-image's map_coordinates will crash if NaN values are passed in
    # coords, so we filter these out (this is a good idea anyway for performance)
    keep = ~np.any(np.isnan(coords), axis=0)

    # At the time of writing, dask-image's map_coordinates prefilter is False
    # by default, we hard-code this here to guard against any changes in
    # default

    output[keep] = dask_image_map_coordinates(
        image, coords[:, keep], prefilter=False, **kwargs
    ).compute()

    reset = np.zeros(coords.shape[1], dtype=bool)

    for i in range(coords.shape[0]):
        reset |= coords[i] < -0.5
        reset |= coords[i] > original_shape[i] - 0.5

    output[reset] = cval

    return output


def map_coordinates(
    image, coords, max_chunk_size=None, output=None, optimize_memory=False, **kwargs
):
    # In the built-in scipy map_coordinates, the values are defined at the
    # center of the pixels. This means that map_coordinates does not
    # correctly treat pixels that are in the outer half of the outer pixels.
    # We solve this by resetting any coordinates that are in the outer half of
    # the border pixels to be at the center of the border pixels. We used to
    # instead pad the array but this was not memory efficient as it ended up
    # producing a copy of the output array.

    # In addition, map_coordinates is not efficient when given big-endian Numpy
    # arrays as it will then make a copy, which is an issue when dealing with
    # memory-mapped FITS files that might be larger than memory. Therefore, for
    # big-endian arrays, we operate in chunks with a size smaller or equal to
    # max_chunk_size.

    # The optimize_memory option isn't used right not by the rest of reproject
    # but it is a mode where if we are in a memory-constrained environment, we
    # re-create memmaps for individual chunks to avoid caching the whole array.
    # We need to decide how to expose this to users.

    # TODO: check how this should behave on a big-endian system.

    from scipy.ndimage import map_coordinates as scipy_map_coordinates

    original_shape = image.shape

    # We copy the coordinates array as we then modify it in-place below to clip
    # to the edges of the array.

    coords = _clip_coords(image, coords)

    # If the data type is native and we are not doing spline interpolation,
    # then scipy_map_coordinates deals properly with memory maps, so we can use
    # it without chunking. Otherwise, we need to iterate over data chunks.
    if image.dtype.isnative and "order" in kwargs and kwargs["order"] <= 1:
        values = scipy_map_coordinates(at_least_float32(image), coords, output=output, **kwargs)
    else:
        if output is None:
            output = np.repeat(np.nan, coords.shape[1])

        values = output

        include = np.ones(coords.shape[1], dtype=bool)

        if "order" in kwargs and kwargs["order"] <= 1:
            padding = 1
        else:
            padding = 10

        for chunk in iterate_chunks(image.shape, max_chunk_size=max_chunk_size):

            include[...] = True
            for idim, slc in enumerate(chunk):
                include[(coords[idim] < slc.start) | (coords[idim] >= slc.stop)] = False

            if not np.any(include):
                continue

            chunk = list(chunk)

            # Adjust chunks to add padding
            for idim, slc in enumerate(chunk):
                start = max(0, slc.start - padding)
                stop = min(original_shape[idim], slc.stop + padding)
                chunk[idim] = slice(start, stop)

            chunk = tuple(chunk)

            coords_subset = coords[:, include].copy()
            for idim, slc in enumerate(chunk):
                coords_subset[idim, :] -= slc.start

            if optimize_memory:
                image_subset = memory_efficient_access(image, chunk)
            else:
                image_subset = image[chunk]

            output[include] = scipy_map_coordinates(
                at_least_float32(image_subset), coords_subset, **kwargs
            )

    reset = np.zeros(coords.shape[1], dtype=bool)

    for i in range(coords.shape[0]):
        reset |= coords[i] < -0.5
        reset |= coords[i] > original_shape[i] - 0.5

    values[reset] = kwargs.get("cval", 0.0)

    return values


def sample_array_edges(shape, *, n_samples):
    # Given an N-dimensional array shape, sample each edge of the array using
    # the requested number of samples (which will include vertices). To do this
    # we iterate through the dimensions and for each one we sample the points
    # in that dimension and iterate over the combination of other vertices.
    # Returns an array with dimensions (N, n_samples)
    all_positions = []
    ndim = len(shape)
    shape = np.array(shape)
    for idim in range(ndim):
        for vertex in range(2**ndim):
            positions = -0.5 + shape * ((vertex & (2 ** np.arange(ndim))) > 0).astype(int)
            positions = np.broadcast_to(positions, (n_samples, ndim)).copy()
            positions[:, idim] = np.linspace(-0.5, shape[idim] - 0.5, n_samples)
            all_positions.append(positions)
    positions = np.unique(np.vstack(all_positions), axis=0).T
    return positions


class ArrayWrapper:

    def __init__(self, array):
        self._array = array
        self.ndim = array.ndim
        self.shape = array.shape
        self.dtype = array.dtype

    def __getitem__(self, item):
        return self._array[item]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
WCS-related utilities
"""

import numpy as np
from astropy.coordinates import SkyCoord, SpectralCoord
from astropy.wcs import WCS
from astropy.wcs.utils import pixel_to_pixel, proj_plane_pixel_scales

__all__ = ["has_celestial", "pixel_to_pixel_with_roundtrip", "pixel_scale"]


def has_celestial(wcs):
    """
    Returns `True` if there are celestial coordinates in the WCS.
    """
    if isinstance(wcs, WCS):
        return wcs.has_celestial
    else:
        for world_axis_class in wcs.low_level_wcs.world_axis_object_classes.values():
            if issubclass(world_axis_class[0], SkyCoord):
                return True
        return False


def has_spectral(wcs):
    """
    Returns `True` if there are spectral coordinates in the WCS.
    """
    if isinstance(wcs, WCS):
        return wcs.has_spectral
    else:
        for world_axis_class in wcs.low_level_wcs.world_axis_object_classes.values():
            if issubclass(world_axis_class[0], SpectralCoord):
                return True
        return False


def pixel_to_pixel_with_roundtrip(wcs1, wcs2, *inputs):
    outputs = pixel_to_pixel(wcs1, wcs2, *inputs)

    # Now convert back to check that coordinates round-trip, if not then set to NaN
    inputs_check = pixel_to_pixel(wcs2, wcs1, *outputs)
    reset = np.zeros(inputs_check[0].shape, dtype=bool)
    for ipix in range(len(inputs_check)):
        reset |= np.abs(inputs_check[ipix] - inputs[ipix]) > 1
    if np.any(reset):
        for ipix in range(len(inputs_check)):
            outputs[ipix] = outputs[ipix].copy()
            outputs[ipix][reset] = np.nan

    return outputs


def pixel_scale(wcs, shape):
    """
    Given a WCS and an image shape, determine the pixel scale.

    If the WCS is an :class:`~astropy.wcs.WCS` instance, the pixel scale is
    the one at the reference coordinate, otherwise it is the scale at the center
    of the image. If the pixels are not square, the smallest pixel scale is
    returned.
    """

    if isinstance(wcs, WCS):
        scales = [
            abs(s) * u
            for (s, u) in zip(proj_plane_pixel_scales(wcs.celestial), wcs.wcs.cunit, strict=False)
        ]
    else:
        # TODO: fix the following for 3D APE-14 WCS
        xp, yp = (shape[1] - 1) / 2, (shape[0] - 1) / 2
        xs = np.array([xp, xp, xp + 1])
        ys = np.array([yp, yp + 1, yp])
        cs = wcs.pixel_to_world(xs, ys)
        scales = (
            abs(cs[0].separation(cs[2])),
            abs(cs[0].separation(cs[1])),
        )

    return min(*scales)



import numpy as np
cimport numpy as np
import cython

ctypedef np.double_t DOUBLE_T

cdef extern from "overlapArea.h":
    double computeOverlap(double * ilon, double * ilat, double * olon, double * olat,
                          int energyMode, double refArea, double * areaRatio)

cdef extern from "reproject_slice_c.h":
    void _reproject_slice_c(int startx, int endx, int starty, int endy, int nx_out, int ny_out,
        double *xp_inout, double *yp_inout, double *xw_in, double *yw_in, double *xw_out, double *yw_out,
        double *array, double *array_new, double *weights,
        int col_in, int col_out, int col_array, int col_new)

# @cython.wraparound(False)
# @cython.boundscheck(False)
def _reproject_slice_cython(int startx, int endx, int starty, int endy, int nx_out, int ny_out,
    np.ndarray[double, ndim=2, mode = "c"] xp_inout,
    np.ndarray[double, ndim=2, mode = "c"] yp_inout,
    np.ndarray[double, ndim=2, mode = "c"] xw_in,
    np.ndarray[double, ndim=2, mode = "c"] yw_in,
    np.ndarray[double, ndim=2, mode = "c"] xw_out,
    np.ndarray[double, ndim=2, mode = "c"] yw_out,
    np.ndarray[double, ndim=2, mode = "c"] array,
    shape_out):

    # Create the array_new and weights objects, plus the objects needed in the loop.
    cdef np.ndarray[double, ndim = 2, mode = "c"] array_new = np.zeros(shape_out, dtype = np.double)
    cdef np.ndarray[double, ndim = 2, mode = "c"] weights = np.zeros(shape_out, dtype = np.double)

    # We need the y size of these 2-dimensional arrays in order to access the elements correctly
    # from raw C.
    cdef int col_in = xw_in.shape[1]
    cdef int col_out = xw_out.shape[1]
    cdef int col_array = array.shape[1]
    cdef int col_new = array_new.shape[1]

    # Call the C function now.
    _reproject_slice_c(startx,endx,starty,endy,nx_out,ny_out,
        &xp_inout[0,0],&yp_inout[0,0],
        &xw_in[0,0],&yw_in[0,0],&xw_out[0,0],&yw_out[0,0],&array[0,0],
        &array_new[0,0],&weights[0,0],
        col_in,col_out,col_array,col_new)

    return array_new,weights

# @cython.wraparound(False)
# @cython.boundscheck(False)
def _compute_overlap(np.ndarray[double, ndim=2] ilon,
                     np.ndarray[double, ndim=2] ilat,
                     np.ndarray[double, ndim=2] olon,
                     np.ndarray[double, ndim=2] olat):
    cdef int i
    cdef int n = ilon.shape[0]

    cdef np.ndarray[double, ndim = 1] overlap = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim = 1] area_ratio = np.empty(n, dtype=np.double)

    for i in range(n):
        overlap[i] = computeOverlap(& ilon[i, 0], & ilat[i, 0], & olon[i, 0], & olat[i, 0],
                                    0, 1, & area_ratio[i])

    return overlap, area_ratio



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings

import numpy as np
from astropy import units as u
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_area

from ._overlap import _reproject_slice_cython


def _reproject_celestial(
    array,
    wcs_in,
    wcs_out,
    shape_out,
    array_out=None,
    output_footprint=None,
    return_footprint=True,
):
    if array_out is None:
        array_out = np.empty(shape_out)

    if output_footprint is None:
        output_footprint = np.empty(shape_out)

    # There may be precision issues at extremely small resolutions due to
    # floating-point limits in WCS coordinate transformations. For more
    # details, see: https://github.com/astropy/reproject/issues/199
    area_threshold = (1e-6 / 3600) ** 2
    if (isinstance(wcs_in, WCS) and proj_plane_pixel_area(wcs_in) < area_threshold) or (
        isinstance(wcs_out, WCS) and proj_plane_pixel_area(wcs_out) < area_threshold
    ):
        warnings.warn(
            "The reproject_exact function may have precision "
            "issues with images that have resolutions below ~1e-6 "
            "arcsec, so the results may not be accurate.",
            UserWarning,
            stacklevel=2,
        )

    # Convert input array to float values. If this comes from a FITS, it might have
    # float32 as value type and that can break things in Cython
    array = np.asarray(array, dtype=float)
    shape_out = tuple(shape_out)

    if wcs_in.pixel_n_dim != 2:
        # TODO: make this work for n-dimensional arrays
        raise NotImplementedError("Only 2-dimensional arrays can be reprojected at this time")
    elif len(shape_out) < wcs_out.low_level_wcs.pixel_n_dim:
        raise ValueError("Too few dimensions in shape_out")
    elif len(array.shape) < wcs_in.low_level_wcs.pixel_n_dim:
        raise ValueError("Too few dimensions in input data")
    elif len(array.shape) != len(shape_out):
        raise ValueError("Number of dimensions in input and output data should match")

    # Separate the "extra" dimensions that don't correspond to a WCS axis and
    # which we'll be looping over
    extra_dimens_in = array.shape[: -wcs_in.low_level_wcs.pixel_n_dim]
    extra_dimens_out = shape_out[: -wcs_out.low_level_wcs.pixel_n_dim]
    if extra_dimens_in != extra_dimens_out:
        raise ValueError("Dimensions to be looped over must match exactly")

    # TODO: at the moment, we compute the coordinates of all of the corners,
    # but we might want to do it in steps for large images.

    # Start off by finding the world position of all the corners of the input
    # image in world coordinates

    ny_in, nx_in = array.shape[-2:]

    x = np.arange(nx_in + 1.0) - 0.5
    y = np.arange(ny_in + 1.0) - 0.5

    xp_in, yp_in = np.meshgrid(x, y, indexing="xy", sparse=False, copy=False)

    world_in = wcs_in.pixel_to_world(xp_in, yp_in)

    # Now compute the world positions of all the corners in the output header

    ny_out, nx_out = shape_out[-2:]

    x = np.arange(nx_out + 1.0) - 0.5
    y = np.arange(ny_out + 1.0) - 0.5

    xp_out, yp_out = np.meshgrid(x, y, indexing="xy", sparse=False, copy=False)

    world_out = wcs_out.pixel_to_world(xp_out, yp_out)

    # Convert the input world coordinates to the frame of the output world
    # coordinates.

    world_in = world_in.transform_to(world_out.frame)

    # Finally, compute the pixel positions in the *output* image of the pixels
    # from the *input* image.

    xp_inout, yp_inout = wcs_out.world_to_pixel(world_in)

    world_in_unitsph = world_in.represent_as("unitspherical")
    xw_in, yw_in = world_in_unitsph.lon.to_value(u.deg), world_in_unitsph.lat.to_value(u.deg)

    world_out_unitsph = world_out.represent_as("unitspherical")
    xw_out, yw_out = world_out_unitsph.lon.to_value(u.deg), world_out_unitsph.lat.to_value(u.deg)

    # If the input array contains extra dimensions beyond what the input WCS
    # has, the extra leading dimensions are assumed to represent multiple
    # images with the same coordinate information. The transformation is
    # computed once and "broadcast" across those images.
    if len(array.shape) == wcs_in.low_level_wcs.pixel_n_dim:
        # We don't need to broadcast the transformation over any extra
        # axes---add an extra axis of length one just so we have something
        # to loop over in all cases.
        broadcasting = False
        array = array.reshape((1, *array.shape))
    elif len(array.shape) > wcs_in.low_level_wcs.pixel_n_dim:
        # We're broadcasting. Flatten the extra dimensions so there's just one
        # to loop over
        broadcasting = True
        array = array.reshape((-1, *array.shape[-wcs_in.low_level_wcs.pixel_n_dim :]))
        array_out = array_out.reshape((-1, *shape_out[-2:]))
        output_footprint = output_footprint.reshape((-1, *shape_out[-2:]))
    else:
        raise ValueError("Too few dimensions for input array")

    array = np.ascontiguousarray(array)

    for i in range(len(array)):
        array_new, weights = _reproject_slice_cython(
            0,
            nx_in,
            0,
            ny_in,
            nx_out,
            ny_out,
            np.ascontiguousarray(xp_inout),
            np.ascontiguousarray(yp_inout),
            np.ascontiguousarray(xw_in),
            np.ascontiguousarray(yw_in),
            np.ascontiguousarray(xw_out),
            np.ascontiguousarray(yw_out),
            array[i],
            shape_out[-2:],
        )

        with np.errstate(invalid="ignore"):
            array_new /= weights

        if broadcasting:
            array_out[i] = array_new
            if return_footprint:
                output_footprint[i] = weights
        else:
            array_out[:] = array_new
            if return_footprint:
                output_footprint[:] = weights

    if broadcasting:
        array_out = array_out.reshape(shape_out)
        if return_footprint:
            output_footprint = output_footprint.reshape(shape_out)

    if return_footprint:
        return array_out, output_footprint
    else:
        return array_out



from astropy.io import fits
from astropy.wcs import WCS
from numpy.testing import assert_allclose

from reproject.conftest import TestLowLevelWCS


def array_footprint_to_hdulist(array, footprint, header):
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU(array, header))
    hdulist.append(fits.ImageHDU(footprint, header, name="footprint"))
    return hdulist


def _underlying_wcs(wcs):
    # For testing purposes, try and return an underlying WCS object if equivalent

    if hasattr(wcs, "low_level_wcs"):
        if isinstance(wcs.low_level_wcs, WCS):
            return wcs.low_level_wcs
        elif isinstance(wcs.low_level_wcs, TestLowLevelWCS):
            return wcs.low_level_wcs._low_level_wcs
    elif isinstance(wcs, TestLowLevelWCS):
        return wcs._low_level_wcs

    return wcs


def assert_wcs_allclose(wcs1, wcs2, **kwargs):
    # First check whether the WCSes are actually the same, either directly
    # or through layers

    if wcs1 is wcs2:
        return True

    if _underlying_wcs(wcs1) is _underlying_wcs(wcs2):
        return True

    header1 = wcs1.to_header()
    header2 = wcs2.to_header()

    assert sorted(header1) == sorted(header2)

    for key1, value1 in header1.items():
        if isinstance(value1, str):
            assert value1 == header2[key1]
        else:
            assert_allclose(value1, header2[key1], **kwargs)



import os
import re
from math import prod

import numpy as np
import pytest
from astropy.coordinates import ICRS
from astropy.io import fits
from astropy.io.fits import Header
from astropy.wcs import WCS
from PIL import Image
from pyavm import AVM

from ... import reproject_interp
from .._high_level import compute_lower_resolution_tiles, reproject_to_hips
from .._trim_utils import fits_getdata_untrimmed, fits_writeto_withtrim
from .._utils import load_properties, tile_header_3d

EXPECTED_FILES = [
    "Norder0/Dir0/Npix0.fits",
    "Norder1/Dir0/Npix2.fits",
    "Norder2/Dir0/Npix9.fits",
    "Norder3/Dir0/Npix38.fits",
    "Norder4/Dir0/Npix152.fits",
    "Norder4/Dir0/Npix153.fits",
    "Norder4/Dir0/Npix154.fits",
    "Norder5/Dir0/Npix609.fits",
    "Norder5/Dir0/Npix611.fits",
    "Norder5/Dir0/Npix612.fits",
    "Norder5/Dir0/Npix614.fits",
    "Norder5/Dir0/Npix617.fits",
    "Norder6/Dir0/Npix2439.fits",
    "Norder6/Dir0/Npix2444.fits",
    "Norder6/Dir0/Npix2445.fits",
    "Norder6/Dir0/Npix2446.fits",
    "Norder6/Dir0/Npix2447.fits",
    "Norder6/Dir0/Npix2450.fits",
    "Norder6/Dir0/Npix2456.fits",
    "Norder6/Dir0/Npix2458.fits",
    "Norder6/Dir0/Npix2468.fits",
    "Norder6/Dir0/Npix2469.fits",
    "index.html",
    "properties",
]


def assert_files_expected(directory, expected):
    actual = sorted(
        [f.relative_to(directory).as_posix() for f in directory.rglob("*") if f.is_file()]
    )
    assert actual == expected


def test_reproject_to_hips(tmp_path, valid_celestial_input_data):

    _, _, input_value, kwargs_in = valid_celestial_input_data

    output_directory = tmp_path / "output"

    reproject_to_hips(
        input_value,
        coord_system_out="equatorial",
        level=6,
        reproject_function=reproject_interp,
        output_directory=output_directory,
        **kwargs_in,
    )

    if str(input_value).endswith("png"):
        expected = [filename.replace(".fits", ".png") for filename in EXPECTED_FILES]
    elif str(input_value).endswith("jpg"):
        expected = [filename.replace(".fits", ".jpg") for filename in EXPECTED_FILES]
    else:
        expected = EXPECTED_FILES

    assert_files_expected(output_directory, expected)


EXPECTED_FILES_GALACTIC = [
    "Norder0/Dir0/Npix9.fits",
    "Norder1/Dir0/Npix39.fits",
    "Norder2/Dir0/Npix156.fits",
    "Norder2/Dir0/Npix157.fits",
    "Norder3/Dir0/Npix627.fits",
    "Norder3/Dir0/Npix630.fits",
    "Norder4/Dir0/Npix2511.fits",
    "Norder4/Dir0/Npix2522.fits",
    "Norder5/Dir10000/Npix10045.fits",
    "Norder5/Dir10000/Npix10047.fits",
    "Norder5/Dir10000/Npix10088.fits",
    "Norder5/Dir10000/Npix10090.fits",
    "Norder6/Dir40000/Npix40182.fits",
    "Norder6/Dir40000/Npix40183.fits",
    "Norder6/Dir40000/Npix40188.fits",
    "Norder6/Dir40000/Npix40189.fits",
    "Norder6/Dir40000/Npix40190.fits",
    "Norder6/Dir40000/Npix40191.fits",
    "Norder6/Dir40000/Npix40354.fits",
    "Norder6/Dir40000/Npix40360.fits",
    "index.html",
    "properties",
]


def test_reproject_to_hips_galactic(tmp_path, simple_celestial_fits_wcs):

    array_in = np.ones((30, 40))
    wcs_in = simple_celestial_fits_wcs

    output_directory = tmp_path / "output"

    reproject_to_hips(
        (array_in, wcs_in),
        coord_system_out="galactic",
        level=6,
        reproject_function=reproject_interp,
        output_directory=output_directory,
    )

    assert_files_expected(output_directory, EXPECTED_FILES_GALACTIC)


def test_reproject_to_hips_invalid_parameters(tmp_path, simple_celestial_fits_wcs):

    array_in = np.ones((30, 40))
    wcs_in = simple_celestial_fits_wcs

    output_directory = tmp_path / "output"

    with pytest.raises(
        ValueError,
        match=re.escape("coord_system_out should be one of equatorial/galactic/ecliptic"),
    ):
        reproject_to_hips(
            (array_in, wcs_in),
            coord_system_out="intragalactic",
            reproject_function=reproject_interp,
            output_directory=output_directory,
        )

    with pytest.raises(ValueError, match=re.escape("tile_size should be even")):
        reproject_to_hips(
            (array_in, wcs_in),
            tile_size=311,
            coord_system_out="galactic",
            reproject_function=reproject_interp,
            output_directory=output_directory,
        )


EXPECTED_FILES_AUTO_1 = [
    "Norder0/Dir0/Npix0.fits",
    "Norder1/Dir0/Npix2.fits",
    "Norder2/Dir0/Npix9.fits",
    "index.html",
    "properties",
]


EXPECTED_FILES_AUTO_2 = [
    "Norder0/Dir0/Npix0.fits",
    "Norder1/Dir0/Npix2.fits",
    "Norder2/Dir0/Npix9.fits",
    "Norder3/Dir0/Npix38.fits",
    "Norder4/Dir0/Npix153.fits",
    "Norder5/Dir0/Npix612.fits",
    "Norder6/Dir0/Npix2450.fits",
    "Norder7/Dir0/Npix9802.fits",
    "index.html",
    "properties",
]


def test_reproject_to_hips_automatic(tmp_path, simple_celestial_wcs):

    array_in = np.ones((30, 40))
    wcs_in = simple_celestial_wcs

    output_directory = tmp_path / "output_1"

    reproject_to_hips(
        (array_in, wcs_in),
        coord_system_out="equatorial",
        reproject_function=reproject_interp,
        output_directory=output_directory,
    )

    assert_files_expected(output_directory, EXPECTED_FILES_AUTO_1)

    if isinstance(wcs_in, WCS):

        output_directory = tmp_path / "output_2"
        wcs_in.wcs.cdelt = -0.001, 0.001

        reproject_to_hips(
            (array_in, wcs_in),
            coord_system_out="equatorial",
            reproject_function=reproject_interp,
            output_directory=output_directory,
        )

        assert_files_expected(output_directory, EXPECTED_FILES_AUTO_2)


def test_reproject_to_hips_alpha(tmp_path, simple_celestial_fits_wcs):

    # Make sure that any input alpha channel is preserved

    data = np.arange(900).reshape((30, 30)) / 1200
    layer = (data * 255).astype(np.uint8)
    alpha = np.zeros((30, 30)).astype(np.uint8)
    alpha[10:15, 10:15] = 255
    array_rgba = np.dstack([layer, np.rot90(layer, k=1), np.rot90(layer, k=2), alpha])
    image = Image.fromarray(array_rgba)
    image.save(tmp_path / "rgb.png")

    avm = AVM.from_wcs(simple_celestial_fits_wcs, shape=(30, 30))
    avm.embed(tmp_path / "rgb.png", tmp_path / "rgb_tagged.png")

    output_directory = tmp_path / "output_1"

    reproject_to_hips(
        tmp_path / "rgb_tagged.png",
        coord_system_out="equatorial",
        reproject_function=reproject_interp,
        output_directory=output_directory,
    )

    result = np.array(Image.open(output_directory / "Norder1" / "Dir0" / "Npix2.png"))

    count = result.sum(axis=(0, 1))

    # There should be many pixels that are valid but transparent
    assert result[:, :, 0].size == 262144
    assert np.all(count[:3] > 50000)
    assert count[3] < 4000


HEADER_SPECTRAL = """
WCSAXES = 3
CRPIX1  = 10.
CRPIX2  = 20.
CRPIX3  = 30.
CDELT1  = -0.02
CDELT2  = 0.02
CDELT3  = 0.1
CUNIT1  = 'deg'
CUNIT2  = 'deg'
CUNIT3  = 'GHz'
CTYPE1  = 'RA---SIN'
CTYPE2  = 'DEC--SIN'
CTYPE3  = 'FREQ'
CRVAL1  = 50.
CRVAL2  = 70.
CRVAL3  = 230.
SPECSYS = 'LSRK'
"""


EXPECTED_FILES_CUBE = [
    "Norder0_7/Dir0_130/Npix0_134.fits",
    "Norder1_8/Dir0_260/Npix3_268.fits",
    "Norder2_9/Dir0_530/Npix15_536.fits",
    "Norder3_10/Dir0_1070/Npix60_1073.fits",
    "Norder4_11/Dir0_2140/Npix240_2147.fits",
    "Norder4_11/Dir0_2140/Npix241_2147.fits",
    "Norder5_12/Dir0_4290/Npix961_4294.fits",
    "Norder5_12/Dir0_4290/Npix964_4294.fits",
    "Norder6_13/Dir0_8580/Npix3845_8588.fits",
    "Norder6_13/Dir0_8580/Npix3845_8589.fits",
    "Norder6_13/Dir0_8580/Npix3856_8588.fits",
    "Norder6_13/Dir0_8580/Npix3856_8589.fits",
    "Norder7_14/Dir10000_17170/Npix15381_17177.fits",
    "Norder7_14/Dir10000_17170/Npix15381_17178.fits",
    "Norder7_14/Dir10000_17170/Npix15381_17179.fits",
    "Norder7_14/Dir10000_17170/Npix15383_17177.fits",
    "Norder7_14/Dir10000_17170/Npix15383_17178.fits",
    "Norder7_14/Dir10000_17170/Npix15383_17179.fits",
    "Norder7_14/Dir10000_17170/Npix15426_17177.fits",
    "Norder7_14/Dir10000_17170/Npix15426_17178.fits",
    "Norder7_14/Dir10000_17170/Npix15426_17179.fits",
    "Norder8_15/Dir60000_34350/Npix61527_34355.fits",
    "Norder8_15/Dir60000_34350/Npix61527_34356.fits",
    "Norder8_15/Dir60000_34350/Npix61527_34357.fits",
    "Norder8_15/Dir60000_34350/Npix61527_34358.fits",
    "Norder8_15/Dir60000_34350/Npix61532_34355.fits",
    "Norder8_15/Dir60000_34350/Npix61532_34356.fits",
    "Norder8_15/Dir60000_34350/Npix61532_34357.fits",
    "Norder8_15/Dir60000_34350/Npix61532_34358.fits",
    "Norder8_15/Dir60000_34350/Npix61533_34355.fits",
    "Norder8_15/Dir60000_34350/Npix61533_34356.fits",
    "Norder8_15/Dir60000_34350/Npix61533_34357.fits",
    "Norder8_15/Dir60000_34350/Npix61533_34358.fits",
    "Norder8_15/Dir60000_34350/Npix61534_34355.fits",
    "Norder8_15/Dir60000_34350/Npix61534_34356.fits",
    "Norder8_15/Dir60000_34350/Npix61534_34357.fits",
    "Norder8_15/Dir60000_34350/Npix61534_34358.fits",
    "Norder8_15/Dir60000_34350/Npix61535_34355.fits",
    "Norder8_15/Dir60000_34350/Npix61535_34356.fits",
    "Norder8_15/Dir60000_34350/Npix61535_34357.fits",
    "Norder8_15/Dir60000_34350/Npix61535_34358.fits",
    "Norder8_15/Dir60000_34350/Npix61704_34355.fits",
    "Norder8_15/Dir60000_34350/Npix61704_34356.fits",
    "Norder8_15/Dir60000_34350/Npix61704_34357.fits",
    "Norder8_15/Dir60000_34350/Npix61704_34358.fits",
    "index.html",
    "properties",
]


def test_reproject_to_hips3d_spectral(tmp_path):

    shape = (15, 16, 17)

    cube_data = np.arange(prod(shape)).reshape(shape)
    cube_wcs = WCS(Header.fromstring(HEADER_SPECTRAL, sep="\n"))

    output_directory = tmp_path / "output"

    reproject_to_hips(
        (cube_data, cube_wcs),
        coord_system_out="equatorial",
        reproject_function=reproject_interp,
        output_directory=output_directory,
        threads=True,
        tile_size=16,
        tile_depth=8,
    )

    assert_files_expected(output_directory, EXPECTED_FILES_CUBE)


def test_properties_regressions(tmp_path, simple_celestial_fits_wcs):
    # Regression tests for the HiPS properties file:
    #  - obs_title should be the dataset directory name, not its parent
    #  - hips_pixel_bitpix should match the float32 tiles that are written
    #  - hips_release_date should follow the ISO 8601 YYYY-mm-ddTHH:MMZ form
    output_directory = tmp_path / "some" / "nested" / "my_survey"

    reproject_to_hips(
        (np.ones((30, 40)), simple_celestial_fits_wcs),
        coord_system_out="equatorial",
        level=3,
        reproject_function=reproject_interp,
        output_directory=output_directory,
    )

    properties = load_properties(output_directory)

    assert properties["obs_title"] == "my_survey"
    assert properties["hips_pixel_bitpix"] == "-32"
    assert re.fullmatch(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}Z", properties["hips_release_date"])

    tile = next(output_directory.glob("Norder3/*/*.fits"))
    assert fits.getheader(tile)["BITPIX"] == -32


def test_trimmed_tiles_store_original_size(tmp_path):
    # Regression test: trimmed 3D tiles must record their original (untrimmed)
    # size via ONAXIS1/2/3 so that a reader can reconstruct the full tile.
    tile_size, tile_depth = 16, 8
    header = tile_header_3d(
        spatial_level=2,
        spatial_index=0,
        spectral_level=2,
        spectral_index=0,
        frame=ICRS(),
        tile_size=tile_size,
        tile_depth=tile_depth,
    )
    if isinstance(header, tuple):
        header = header[0]
    crpix_before = (header["CRPIX1"], header["CRPIX2"], header["CRPIX3"])

    array = np.full((tile_depth, tile_size, tile_size), np.nan)
    array[2:5, 3:10, 4:12] = 1.0  # off-centre block so every axis is trimmed

    filename = str(tmp_path / "tile.fits")
    fits_writeto_withtrim(filename, array, header)

    written = fits.getheader(filename)
    # The original full size is preserved in ONAXISn ...
    assert (written["ONAXIS1"], written["ONAXIS2"], written["ONAXIS3"]) == (
        tile_size,
        tile_size,
        tile_depth,
    )
    # ... while NAXISn and the reference pixel reflect the trimmed, stored data
    assert (written["NAXIS1"], written["NAXIS2"], written["NAXIS3"]) == (8, 7, 3)
    assert (written["TRIM1"], written["TRIM2"], written["TRIM3"]) == (4, 3, 2)
    # The reference pixel on each axis is shifted by the leading trim
    assert written["CRPIX1"] == crpix_before[0] - 4
    assert written["CRPIX2"] == crpix_before[1] - 3
    assert written["CRPIX3"] == crpix_before[2] - 2

    # The reader restores the full tile, relying on ONAXISn even without hints
    restored = fits_getdata_untrimmed(filename, tile_size=1, tile_depth=1)
    assert restored.shape == (tile_depth, tile_size, tile_size)
    np.testing.assert_array_equal(np.nan_to_num(restored), np.nan_to_num(array.astype(np.float32)))


CUBE_CLAMP_HEADER = """
WCSAXES = 3
CRPIX1  = 16.
CRPIX2  = 16.
CRPIX3  = 1.
CDELT1  = -0.15
CDELT2  = 0.15
CDELT3  = {cdelt3}
CUNIT1  = 'deg'
CUNIT2  = 'deg'
CUNIT3  = 'Hz'
CTYPE1  = 'RA---SIN'
CTYPE2  = 'DEC--SIN'
CTYPE3  = 'FREQ-LOG'
CRVAL1  = 50.
CRVAL2  = 70.
CRVAL3  = 1.0E9
SPECSYS = 'LSRK'
"""


def test_reproject_to_hips3d_clamps_spectral_order(tmp_path):
    # Regression test: when the spectral (Lmax) order is smaller than the
    # spatial (Kmax) order, lower-resolution tiles must clamp the spectral
    # order at 0 (never negative) and keep degrading spatially.
    #
    # This needs Lmax < Kmax, which means a low spectral order. Because the
    # frequency grid spans FREQ_MIN=1e-18 to FREQ_MAX=1e38 Hz, a low order only
    # arises for a band covering many decades, so we use a coarse spatial grid
    # (small Kmax) together with a wide, log-sampled spectral axis (small Lmax).
    level, level_depth = 4, 1  # Kmax > Lmax forces the clamp

    nz = 8
    # CDELT3 for the -LOG algorithm so the nz channels span ~4 decades
    cdelt3 = 1e9 * np.log(10**4) / (nz - 1)
    header = Header.fromstring(CUBE_CLAMP_HEADER.format(cdelt3=cdelt3), sep="\n")
    cube = np.arange(nz * 32 * 32).reshape(nz, 32, 32).astype(float)

    output_directory = tmp_path / "cube"

    reproject_to_hips(
        (cube, WCS(header)),
        coord_system_out="equatorial",
        reproject_function=reproject_interp,
        output_directory=output_directory,
        tile_size=16,
        tile_depth=8,
        level=level,
        level_depth=level_depth,
    )

    pairs = sorted(
        tuple(int(value) for value in path.name.replace("Norder", "").split("_"))
        for path in output_directory.glob("Norder*")
    )

    # Orders follow L = max(0, Lmax - (Kmax - K)), clamped at 0, down to spatial 0
    expected = sorted({(level - sub, max(0, level_depth - sub)) for sub in range(level + 1)})
    assert pairs == expected
    assert all(spectral_order >= 0 for _, spectral_order in pairs)
    assert min(spatial_order for spatial_order, _ in pairs) == 0
    assert min(spectral_order for _, spectral_order in pairs) == 0

    # Clamped lower-resolution tiles still carry the propagated data
    clamped = list(output_directory.glob("Norder0_0/*/*.fits"))
    assert clamped
    assert np.isfinite(fits.getdata(clamped[0])).any()


def test_compute_lower_resolution_no_tiles_warns(tmp_path):
    # Regression test: if no tiles were generated at the deepest level, the
    # lower-resolution step should warn and return rather than raise.
    output_directory = tmp_path / "empty"
    os.makedirs(output_directory / "Norder4_2")

    with pytest.warns(UserWarning, match="No tiles were generated"):
        compute_lower_resolution_tiles(
            output_directory=output_directory,
            ndim=3,
            frame=ICRS(),
            tile_format="fits",
            tile_size=8,
            tile_depth=4,
            spatial_level=4,
            level_depth=2,
        )

    assert list(output_directory.glob("Norder*")) == [output_directory / "Norder4_2"]


# TODO: Add tests of different spectral frames
# TODO: Add test of auto level determination for all sky maps in 2D and 3D



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from itertools import product

import numpy as np
import pytest
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS
from astropy.wcs.wcsapi import HighLevelWCSWrapper, SlicedLowLevelWCS
from numpy.testing import assert_allclose

from ...tests.helpers import array_footprint_to_hdulist
from .._high_level import reproject_adaptive


def as_high_level_wcs(wcs):
    return HighLevelWCSWrapper(SlicedLowLevelWCS(wcs, Ellipsis))


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.array_compare(single_reference=True)
@pytest.mark.parametrize("wcsapi", (False, True))
@pytest.mark.parametrize("center_jacobian", (False, True))
@pytest.mark.parametrize("roundtrip_coords", (False, True))
def test_reproject_adaptive_2d(wcsapi, center_jacobian, roundtrip_coords):
    # Set up initial array with pattern
    data_in = np.zeros((256, 256))
    data_in[::20, :] = 1
    data_in[:, ::20] = 1
    data_in[10::20, 10::20] = 1

    # Define a simple input WCS
    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = 128.5, 128.5
    wcs_in.wcs.cdelt = -0.01, 0.01

    # Define a lower resolution output WCS
    wcs_out = WCS(naxis=2)
    wcs_out.wcs.crpix = 30.5, 30.5
    wcs_out.wcs.cdelt = -0.0427, 0.0427

    header_out = wcs_out.to_header()

    if wcsapi:  # Enforce a pure wcsapi API
        wcs_in = as_high_level_wcs(wcs_in)
        wcs_out = as_high_level_wcs(wcs_out)

    array_out, footprint_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(60, 60),
        center_jacobian=center_jacobian,
        roundtrip_coords=roundtrip_coords,
        kernel="hann",
        boundary_mode="ignore",
    )

    # Check that surface brightness is conserved in the unrotated case
    assert_allclose(np.nansum(data_in), np.nansum(array_out) * (256 / 60) ** 2, rtol=0.1)

    return array_footprint_to_hdulist(array_out, footprint_out, header_out)


def test_reproject_adaptive_dtypes():
    # Set up initial array with pattern
    data_in = np.zeros((256, 256))
    data_in[::20, :] = 1
    data_in[:, ::20] = 1
    data_in[10::20, 10::20] = 1

    # Define a simple input WCS
    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = 128.5, 128.5
    wcs_in.wcs.cdelt = -0.01, 0.01

    # Define a lower resolution output WCS
    wcs_out = WCS(naxis=2)
    wcs_out.wcs.crpix = 30.5, 30.5
    wcs_out.wcs.cdelt = -0.0427, 0.0427

    # Run with double-precision inputs and outputs
    array_out, footprint_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(60, 60),
    )

    # Run with single precision
    array_out_32bit = np.empty_like(array_out, dtype=np.float32)
    _, footprint_out = reproject_adaptive(
        (data_in.astype(np.float32), wcs_in),
        wcs_out,
        shape_out=(60, 60),
        output_array=array_out_32bit,
    )

    # Check that the results are similar
    assert_allclose(array_out, array_out_32bit, rtol=1e-6)


@pytest.mark.parametrize("axis", ("x", "y"))
@pytest.mark.parametrize("center_jacobian", (True, False))
def test_reproject_adaptive_despike_jacobian(axis, center_jacobian):
    if axis == "x":
        wcs_in = WCS(naxis=2)
        wcs_in.wcs.ctype = "RA---CAR", "DEC--CAR"
        wcs_in.wcs.crpix = 18.5, 9.5
        wcs_in.wcs.crval = 180, 0
        wcs_in.wcs.cdelt = 10, 10
        data_in = np.arange(18 * 36).reshape((18, 36))

        wcs_out = WCS(naxis=2)
        wcs_out.wcs.ctype = "RA---CAR", "DEC--CAR"
        wcs_out.wcs.crpix = 20, 20
        wcs_out.wcs.crval = 10, 0
        wcs_out.wcs.cdelt = 5, 1
    else:
        wcs_in = WCS(naxis=2)
        wcs_in.wcs.ctype = "DEC--CAR", "RA---CAR"
        wcs_in.wcs.crpix = 9.5, 18.5
        wcs_in.wcs.crval = 0, 180
        wcs_in.wcs.cdelt = 10, 10
        data_in = np.arange(18 * 36).reshape((36, 18))

        wcs_out = WCS(naxis=2)
        wcs_out.wcs.ctype = "DEC--CAR", "RA---CAR"
        wcs_out.wcs.crpix = 20, 20
        wcs_out.wcs.crval = 0, 10
        wcs_out.wcs.cdelt = 1, 5

    output = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        (40, 40),
        roundtrip_coords=False,
        return_footprint=False,
        despike_jacobian=False,
        center_jacobian=center_jacobian,
        x_cyclic=(axis == "x"),
        y_cyclic=(axis == "y"),
    )

    output_despiked = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        (40, 40),
        roundtrip_coords=False,
        return_footprint=False,
        despike_jacobian=True,
        center_jacobian=center_jacobian,
        x_cyclic=(axis == "x"),
        y_cyclic=(axis == "y"),
    )

    finite = np.isfinite(output)
    # There should be a strip of nans in the un-despiked output
    assert np.sum(finite) < output.size
    # The outputs should match where spiking wasn't a concern
    np.testing.assert_equal(output[finite], output_despiked[finite])
    # The despiked output should have no nans
    assert np.all(np.isfinite(output_despiked))


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.array_compare(single_reference=True)
@pytest.mark.parametrize("center_jacobian", (False, True))
@pytest.mark.parametrize("roundtrip_coords", (False, True))
def test_reproject_adaptive_2d_rotated(center_jacobian, roundtrip_coords):
    # Set up initial array with pattern
    data_in = np.zeros((256, 256))
    data_in[::20, :] = 1
    data_in[:, ::20] = 1
    data_in[10::20, 10::20] = 1

    # Define a simple input WCS
    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = 128.5, 128.5
    wcs_in.wcs.cdelt = -0.01, 0.01

    # Define a lower resolution output WCS with rotation
    wcs_out = WCS(naxis=2)
    wcs_out.wcs.crpix = 30.5, 30.5
    wcs_out.wcs.cdelt = -0.0427, 0.0427
    wcs_out.wcs.pc = [[0.8, 0.2], [-0.2, 0.8]]

    header_out = wcs_out.to_header()

    array_out, footprint_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(60, 60),
        center_jacobian=center_jacobian,
        roundtrip_coords=roundtrip_coords,
        kernel="hann",
        boundary_mode="ignore",
    )

    return array_footprint_to_hdulist(array_out, footprint_out, header_out)


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.parametrize("center_jacobian", (False, True))
@pytest.mark.parametrize("kernel", ("hann", "gaussian"))
def test_reproject_adaptive_high_aliasing_potential_rotation(
    roundtrip_coords, center_jacobian, kernel
):
    # Generate sample data with vertical stripes alternating with every column
    data_in = np.arange(40 * 40).reshape((40, 40))
    data_in = (data_in) % 2

    # Set up the input image coordinates, defining pixel coordinates as world
    # coordinates (with an offset)
    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = 21, 21
    wcs_in.wcs.crval = 0, 0
    wcs_in.wcs.cdelt = 1, 1

    # Set up the output image coordinates
    wcs_out = WCS(naxis=2)
    wcs_out.wcs.crpix = 3, 5
    wcs_out.wcs.crval = 0, 0
    wcs_out.wcs.cdelt = 2, 1

    array_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(11, 6),
        return_footprint=False,
        roundtrip_coords=roundtrip_coords,
        center_jacobian=center_jacobian,
        kernel=kernel,
    )

    # The CDELT1 value in wcs_out produces a down-sampling by a factor of two
    # along the output x axis. With the input image containing vertical lines
    # with values of zero or one, we should have uniform values of 0.5
    # throughout our output array.
    np.testing.assert_allclose(array_out, 0.5, rtol=0.002)

    # Within the transforms, the order of operations is:
    # input pixel coordinates -> input rotation -> input scaling
    # -> world coords -> output scaling -> output rotation
    # -> output pixel coordinates. So if we add a 90-degree rotation to the
    # output image, we're declaring that image-x is world-y and vice-versa, but
    # since we're rotating the already-downsampled image, no pixel values
    # should change
    angle = 90 * np.pi / 180
    wcs_out.wcs.pc = [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
    array_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(11, 6),
        return_footprint=False,
        roundtrip_coords=roundtrip_coords,
        center_jacobian=center_jacobian,
        kernel=kernel,
    )
    np.testing.assert_allclose(array_out, 0.5, rtol=0.002)

    # But if we add a 90-degree rotation to the input coordinates, then when
    # our stretched output pixels are projected onto the input data, they will
    # be stretched along the stripes, rather than perpendicular to them, and so
    # we'll still see the alternating stripes in our output data---whether or
    # not wcs_out contains a rotation.
    # For these last two cases, we only use a Hann kernel, since the blurring
    # inherent with the Gaussian kernel makes the comparison more difficult.
    angle = 90 * np.pi / 180
    wcs_in.wcs.pc = [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
    array_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(11, 6),
        return_footprint=False,
        roundtrip_coords=roundtrip_coords,
        center_jacobian=center_jacobian,
        kernel="hann",
    )

    # Generate the expected pattern of alternating stripes
    data_ref = np.arange(array_out.shape[1]) % 2
    data_ref = np.vstack([data_ref] * array_out.shape[0])
    np.testing.assert_allclose(array_out, data_ref)

    # Clear the rotation in the output coordinates
    wcs_out.wcs.pc = [[1, 0], [0, 1]]
    array_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(11, 6),
        return_footprint=False,
        roundtrip_coords=roundtrip_coords,
        center_jacobian=center_jacobian,
        kernel="hann",
    )
    data_ref = np.arange(array_out.shape[0]) % 2
    data_ref = np.vstack([data_ref] * array_out.shape[1]).T
    np.testing.assert_allclose(array_out, data_ref)


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.parametrize("center_jacobian", (False, True))
def test_reproject_adaptive_high_aliasing_potential_shearing(roundtrip_coords, center_jacobian):
    # Generate sample data with vertical stripes alternating with every column
    data_in = np.arange(40 * 40).reshape((40, 40))
    data_in = (data_in) % 2

    # Set up the input image coordinates, defining pixel coordinates as world
    # coordinates (with an offset)
    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = 21, 21
    wcs_in.wcs.crval = 0, 0
    wcs_in.wcs.cdelt = 1, 1

    for shear_x in (-1, 0, 1):
        for shear_y in (-1, 0, 1):
            if shear_x == shear_y == 0:
                continue

            # Set up the output image coordinates, with shearing in both x
            # and y
            wcs_out = WCS(naxis=2)
            wcs_out.wcs.crpix = 3, 5
            wcs_out.wcs.crval = 0, 0
            wcs_out.wcs.cdelt = 1, 1
            wcs_out.wcs.pc = np.array([[1, shear_x], [0, 1]]) @ np.array([[1, 0], [shear_y, 1]])

            # n.b. The Gaussian kernel is much better-behaved in this
            # particular scenario, so using it allows a much smaller tolerance
            # in the comparison. Likewise, the kernel width is boosted to 1.5
            # to achieve stronger anti-aliasing for this test.
            array_out = reproject_adaptive(
                (data_in, wcs_in),
                wcs_out,
                shape_out=(11, 6),
                return_footprint=False,
                roundtrip_coords=False,
                center_jacobian=center_jacobian,
                kernel="gaussian",
                kernel_width=1.5,
            )

            # We should get values close to 0.5 (an average of the 1s and 0s
            # in the input image). This is as opposed to values near 0 or 1,
            # which would indicate incorrect averaging of sampled points.
            np.testing.assert_allclose(array_out, 0.5, atol=0.02, rtol=0)


@pytest.mark.filterwarnings("ignore::FutureWarning")
def test_reproject_adaptive_flux_conservation():
    # This is more than just testing the `conserve_flux` flag---the expectation
    # that flux should be conserved gives a very easy way to quantify how
    # correct an image is. We use that to run through a grid of affine
    # transformations, checking that the output is correct for each one.

    # Generate input data of a single point---this is the most unforgiving
    # setup, as there's no change for multiple pixels to average out if their
    # reprojected flux goes slightly above or below being conserved.
    data_in = np.zeros((20, 20))
    data_in[12, 13] = 1

    # Set up the input image coordinates, defining pixel coordinates as world
    # coordinates (with an offset)
    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = 11, 11
    wcs_in.wcs.crval = 0, 0
    wcs_in.wcs.cdelt = 1, 1

    wcs_out = WCS(naxis=2)
    wcs_out.wcs.crpix = 16, 16
    wcs_out.wcs.crval = 0, 0

    # Define our grid of affine transformations. Even with only a few values
    # for each parameter, this is by far the longest-running test in this file,
    # so it's tough to justify a finer grid.
    rotations = [0, 45, 80, 90]
    scales_x = np.logspace(-0.2, 0.2, 3)
    scales_y = np.logspace(-0.3, 0.3, 3)
    translations_x = np.linspace(0, 8, 3)
    translations_y = np.linspace(0, 0.42, 3)
    shears_x = np.linspace(-0.7, 0.7, 3)
    shears_y = np.linspace(-0.2, 0.2, 3)

    for rot, scale_x, scale_y, _, _, shear_x, shear_y in product(
        rotations, scales_x, scales_y, translations_x, translations_y, shears_x, shears_y
    ):
        wcs_out.wcs.cdelt = scale_x, scale_y
        angle = rot * np.pi / 180
        rot_matrix = np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])
        shear_x_matrix = np.array([[1, shear_x], [0, 1]])
        shear_y_matrix = np.array([[1, 0], [shear_y, 1]])
        wcs_out.wcs.pc = rot_matrix @ shear_x_matrix @ shear_y_matrix

        # The Gaussian kernel does a better job at flux conservation, so
        # choosing it here allows a tighter tolerance. Increasing the sample
        # region width also allows a tighter tolerance---less room for bugs to
        # hide! Setting the boundary mode to 'ignore' avoids excessive clipping
        # of the edges of our small output image, letting us use a smaller
        # image and do more tests in the same time.
        array_out = reproject_adaptive(
            (data_in, wcs_in),
            wcs_out,
            shape_out=(30, 30),
            return_footprint=False,
            roundtrip_coords=False,
            center_jacobian=False,
            kernel="gaussian",
            sample_region_width=5,
            conserve_flux=True,
            boundary_mode="ignore",
        )

        # The degree of flux-conservation we end up seeing isn't necessarily
        # something that we can constrain a priori, so here we test for
        # conservation to within 0.4% because that's as good as it's getting
        # right now. This test is more about ensuring future bugs don't
        # accidentally worsen flux conservation than ensuring that some
        # required threshold is met.
        np.testing.assert_allclose(np.nansum(array_out), 1, rtol=0.004)


@pytest.mark.parametrize("x_cyclic", (False, True))
@pytest.mark.parametrize("y_cyclic", (False, True))
@pytest.mark.parametrize("rotated", (False, True))
def test_boundary_modes(x_cyclic, y_cyclic, rotated):
    # Vertical-stripe test data
    data_in = np.arange(30) % 2
    data_in = np.vstack([data_in] * 30)

    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = [15.5, 15.5]
    wcs_in.wcs.crval = [0, 0]
    wcs_in.wcs.cdelt = [1, 1]

    # We'll center the input image in the larger output canvas, but otherwise
    # we'll do nothing. Except in the rotated case, where we'll rotate the
    # input array by transposing it and counter-rotate the output WCS, so that
    # the output should be the same.
    wcs_out = wcs_in.deepcopy()
    wcs_out.wcs.crpix = [20.5, 20.5]

    if rotated:
        data_in = data_in.T
        angle = 90 * np.pi / 180
        wcs_out.wcs.pc = [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
        y_appears_cyclic = x_cyclic
        x_appears_cyclic = y_cyclic
    else:
        x_appears_cyclic = x_cyclic
        y_appears_cyclic = y_cyclic

    # We'll step through each boundary mode and check its output.
    # For all boundary modes, the outermost rows and columns should show the
    # repeating, anti-aliased stripe pattern if the corresponding axis is
    # cyclic. This is tested by comparing every other column to an expected
    # value pulled from an appropriate pixel in the interior of the output
    # image.
    # The sample region width is set to 4.00001 so that the sample region
    # boundary doesn't fall exactly on input pixel centers. Otherwise,
    # floating-point error can cause pixel-to-pixel variation in whether an
    # input sample is just-barely-included or just-barely-excluded, especially
    # in the rotated case.

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="strict",
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )
    # We've placed a 30x30 input image at the center of a 40x40 output image.
    # With strict boundaries, we should get NaNs for the 5 rows and columns on
    # each side of the output image, being outside the bounds of the input
    # image. The NaNs should also span an additional two rows/columns due
    # to the width of the Gaussian kernel.
    if y_appears_cyclic:
        np.testing.assert_allclose(data_out[:, 7:-7:2], data_out[15, 11])
        np.testing.assert_allclose(data_out[:, 8:-7:2], data_out[15, 10])
    else:
        assert np.all(np.isnan(data_out[:7]))
        assert np.all(np.isnan(data_out[-7:]))

    if x_appears_cyclic:
        np.testing.assert_allclose(data_out[7:-7, ::2], data_out[15, 10])
        np.testing.assert_allclose(data_out[7:-7, 1::2], data_out[15, 11])
    else:
        assert np.all(np.isnan(data_out[:, :7]))
        assert np.all(np.isnan(data_out[:, -7:]))

    assert not np.any(np.isnan(data_out[7:-7, 7:-7]))

    # The 'ignore_threshold' mode should match the 'strict' mode when the
    # threshold is set to 0.
    data_out_ignore_threshold = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="ignore_threshold",
        boundary_ignore_threshold=0,
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )

    np.testing.assert_array_equal(data_out, data_out_ignore_threshold)

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="constant",
        boundary_fill_value=1000,
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )
    # With constant filling, we should get NaNs for the
    # outermost 3 rows/columns. The next four rows/columns should have very
    # large values due to the kernel width and our fill value. The remaining
    # values should be within the [0, 1] range of the input data.
    if y_appears_cyclic:
        np.testing.assert_allclose(data_out[:, 7:-7:2], data_out[15, 11])
        np.testing.assert_allclose(data_out[:, 8:-7:2], data_out[15, 10])
    else:
        assert np.all(np.isnan(data_out[:3]))
        np.testing.assert_array_less(1, data_out[3:7, 3:-3])
        assert np.all(np.isnan(data_out[-3:]))
        np.testing.assert_array_less(1, data_out[-7:-3, 3:-3])

    if x_appears_cyclic:
        np.testing.assert_allclose(data_out[7:-7, ::2], data_out[15, 10])
        np.testing.assert_allclose(data_out[7:-7, 1::2], data_out[15, 11])
    else:
        assert np.all(np.isnan(data_out[:, :3]))
        np.testing.assert_array_less(1, data_out[3:-3, 3:7])
        assert np.all(np.isnan(data_out[:, -3:]))
        np.testing.assert_array_less(1, data_out[3:-3, -7:-3])

    assert np.all(data_out[7:-7, 7:-7] <= 1)

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="grid-constant",
        boundary_fill_value=1000,
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )
    # With grid-constant filling, we should get our fill value for the
    # outermost 3 rows/columns. The next four rows/columns should have very
    # large values due to the kernel width and our fill value. The remaining
    # values should be within the [0, 1] range of the input data. There should
    # be no NaNs.
    if y_appears_cyclic:
        np.testing.assert_allclose(data_out[:, 7:-7:2], data_out[15, 11])
        np.testing.assert_allclose(data_out[:, 8:-7:2], data_out[15, 10])
    else:
        np.testing.assert_equal(data_out[:3], 1000)
        np.testing.assert_array_less(1, data_out[3:7])
        np.testing.assert_equal(data_out[-3:], 1000)
        np.testing.assert_array_less(1, data_out[-7:-3])

    if x_appears_cyclic:
        np.testing.assert_allclose(data_out[7:-7, ::2], data_out[15, 10])
        np.testing.assert_allclose(data_out[7:-7, 1::2], data_out[15, 11])
    else:
        np.testing.assert_equal(data_out[:, :3], 1000)
        np.testing.assert_array_less(1, data_out[:, 3:7])
        np.testing.assert_equal(data_out[:, -3:], 1000)
        np.testing.assert_array_less(1, data_out[:, -7:-3])

    assert np.all(data_out[7:-7, 7:-7] <= 1)
    assert not np.any(np.isnan(data_out))

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="ignore",
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )
    # With missing samples ignored, we should get NaNs for three pixels on all
    # edges, and no NaNs elsewhere. The next four rows on top and bottom, where
    # the sampling region spans the boundary, should match the interior of the
    # image to within floating-point error, due to the vertical symmetry of the
    # image. Along the sides, there should be three columns of near-zero on the
    # left and near-one on the right, as those rows only or primarily sample
    # the edge column.
    if y_appears_cyclic:
        np.testing.assert_allclose(data_out[:, 7:-7:2], data_out[15, 11])
        np.testing.assert_allclose(data_out[:, 8:-7:2], data_out[15, 10])
    else:
        assert np.all(np.isnan(data_out[:3]))
        assert np.all(np.isnan(data_out[-3:]))
        np.testing.assert_allclose(data_out[3:7, 3:-3], data_out[7:11, 3:-3])
        np.testing.assert_allclose(data_out[-7:-3, 3:-3], data_out[7:11, 3:-3])

    if x_appears_cyclic:
        np.testing.assert_allclose(data_out[7:-7, ::2], data_out[15, 10])
        np.testing.assert_allclose(data_out[7:-7, 1::2], data_out[15, 11])
    else:
        assert np.all(np.isnan(data_out[:, :3]))
        assert np.all(np.isnan(data_out[:, -3:]))
        np.testing.assert_array_less(data_out[3:-3, 3:6], 0.5)
        np.testing.assert_array_less(0.5, data_out[3:-3, -6:-3])

    assert not np.any(np.isnan(data_out[3:-3, 3:-3]))

    # The 'ignore_threshold' mode should match the 'ignore' mode when the
    # threshold is set to 1.
    data_out_ignore_threshold = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="ignore_threshold",
        boundary_ignore_threshold=1,
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )

    np.testing.assert_array_equal(data_out, data_out_ignore_threshold)

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="ignore_threshold",
        boundary_ignore_threshold=0.5,
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )
    # With the threshold set to 0.5, 'ignore_threshold' should look like the
    # 'ignore' output, except that the band of NaNs along the outside is 5
    # pixels thick instead of 3---an extra 2 rows and columns are set to NaN
    # because those output pixels sample mostly non-existent input pixels.
    if y_appears_cyclic:
        np.testing.assert_allclose(data_out[:, 7:-7:2], data_out[15, 11])
        np.testing.assert_allclose(data_out[:, 8:-7:2], data_out[15, 10])
    else:
        assert np.all(np.isnan(data_out[:5]))
        assert np.all(np.isnan(data_out[-5:]))
        np.testing.assert_allclose(data_out[5:7, 5:-5], data_out[7:9, 5:-5])
        np.testing.assert_allclose(data_out[-7:-5, 5:-5], data_out[7:9, 5:-5])

    if x_appears_cyclic:
        np.testing.assert_allclose(data_out[7:-7, ::2], data_out[15, 10])
        np.testing.assert_allclose(data_out[7:-7, 1::2], data_out[15, 11])
    else:
        assert np.all(np.isnan(data_out[:, :5]))
        assert np.all(np.isnan(data_out[:, -5:]))
        np.testing.assert_array_less(data_out[5:-5, 5:6], 0.5)
        np.testing.assert_array_less(0.5, data_out[5:-5, -6:-5])

    assert not np.any(np.isnan(data_out[5:-5, 5:-5]))

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(40, 40),
        return_footprint=False,
        kernel="gaussian",
        sample_region_width=4.00001,
        boundary_mode="nearest",
        x_cyclic=x_cyclic,
        y_cyclic=y_cyclic,
    )
    # With out-of-bounds samples replaced with the nearest valid sample, the
    # top and bottom rows should be the same as the central rows. There should
    # be three columns of 0 on the left and 1 on the right, and the next three
    # columns should be small or large, respectively, since they're sampling
    # lots of repeated 0s or 1s. There should be no NaNs.
    if y_appears_cyclic:
        np.testing.assert_allclose(data_out[:, 7:-7:2], data_out[15, 11])
        np.testing.assert_allclose(data_out[:, 8:-7:2], data_out[15, 10])
    else:
        np.testing.assert_allclose(data_out[:7], data_out[7:14])
        np.testing.assert_allclose(data_out[-7:], data_out[7:14])

    if x_appears_cyclic:
        np.testing.assert_allclose(data_out[7:-7, ::2], data_out[15, 10])
        np.testing.assert_allclose(data_out[7:-7, 1::2], data_out[15, 11])
    else:
        np.testing.assert_allclose(data_out[:, :3], 0)
        np.testing.assert_allclose(data_out[:, -3:], 1)
        np.testing.assert_array_less(data_out[:, 3:6], 0.5)
        np.testing.assert_array_less(0.5, data_out[:, -6:-3])

    assert not np.any(np.isnan(data_out))


@pytest.mark.parametrize("bad_val", (np.nan, np.inf))
def test_bad_value_modes(bad_val):
    bad_val_checker = np.isnan if np.isnan(bad_val) else np.isinf
    data_in = np.ones((30, 30))
    data_in[15, 15] = bad_val

    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = [15.5, 15.5]
    wcs_in.wcs.crval = [0, 0]
    wcs_in.wcs.cdelt = [1, 1]

    # Our reprojection will be a no-op, so it's easy to reason about what
    # should happen
    wcs_out = wcs_in.deepcopy()

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(30, 30),
        return_footprint=False,
        boundary_mode="ignore",
        sample_region_width=3,
        bad_value_mode="strict",
    )

    # With a sample_region_width of 3, we expect a 3x3 box of nans centered on
    # the input-image nan.
    assert np.all(bad_val_checker(data_out[14:17, 14:17]))
    # And they should be the only nans
    assert np.sum(bad_val_checker(data_out)) == 9

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(30, 30),
        return_footprint=False,
        boundary_mode="ignore",
        sample_region_width=3,
        bad_value_mode="constant",
        bad_fill_value=10,
    )

    # Now, where there were nans, we should get values > 1, since 10 is our
    # fill value.
    assert np.all(data_out[14:17, 14:17] > 1)
    data_out[14:17, 14:17] = 1
    np.testing.assert_equal(data_out, 1)

    data_out = reproject_adaptive(
        (data_in, wcs_in),
        wcs_out,
        shape_out=(30, 30),
        return_footprint=False,
        boundary_mode="ignore",
        sample_region_width=3,
        bad_value_mode="ignore",
    )

    # Now the nan should be ignored and we should only get 1s coming out
    np.testing.assert_equal(data_out, 1)


def test_invald_bad_value_mode():
    data_in = np.ones((30, 30))

    wcs_in = WCS(naxis=2)
    wcs_in.wcs.crpix = [15.5, 15.5]
    wcs_in.wcs.crval = [0, 0]
    wcs_in.wcs.cdelt = [1, 1]

    wcs_out = wcs_in.deepcopy()

    with pytest.raises(ValueError, match="bad_value_mode 'invalid_mode' not recognized"):
        reproject_adaptive(
            (data_in, wcs_in),
            wcs_out,
            shape_out=(30, 30),
            return_footprint=False,
            boundary_mode="ignore",
            sample_region_width=3,
            bad_value_mode="invalid_mode",
        )


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.array_compare(single_reference=True)
def test_reproject_adaptive_roundtrip(aia_test_data):
    # Test the reprojection with solar data, which ensures that the masking of
    # pixels based on round-tripping works correctly. Using asdf is not just
    # about testing a different format but making sure that GWCS works.

    data, wcs, target_wcs = aia_test_data

    output, footprint = reproject_adaptive(
        (data, wcs), target_wcs, (128, 128), center_jacobian=True, kernel="hann"
    )

    header_out = target_wcs.to_header()

    header_out["DATE-OBS"] = header_out["DATE-OBS"].replace("T", " ")

    # With sunpy 6.0.0 and later, additional keyword arguments are written out
    # so we remove these as they are not important for the comparison with the
    # reference files.
    header_out.pop("DATE-AVG", None)
    header_out.pop("MJD-AVG", None)

    return array_footprint_to_hdulist(output, footprint, header_out)


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.array_compare(single_reference=True)
def test_reproject_adaptive_uncentered_jacobian(aia_test_data):
    # Explicitly test the uncentered-Jacobian path for a non-affine transform.
    # For this case, output pixels change by 6% at most, and usually much less.
    # (Though more nan pixels are present, as the uncentered calculation draws
    # in values from a bit further away.)

    data, wcs, target_wcs = aia_test_data

    output, footprint = reproject_adaptive(
        (data, wcs), target_wcs, (128, 128), center_jacobian=False, kernel="hann"
    )

    header_out = target_wcs.to_header()

    header_out["DATE-OBS"] = header_out["DATE-OBS"].replace("T", " ")

    # With sunpy 6.0.0 and later, additional keyword arguments are written out
    # so we remove these as they are not important for the comparison with the
    # reference files.
    header_out.pop("DATE-AVG", None)
    header_out.pop("MJD-AVG", None)

    return array_footprint_to_hdulist(output, footprint, header_out)


def _setup_for_broadcast_test(
    conserve_flux=False,
    boundary_mode="strict",
    kernel="gaussian",
    center_jacobian=False,
    bad_value_mode="strict",
):
    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:
        hdu_in = pf[0]
        header_in = hdu_in.header.copy()
        header_out = hdu_in.header.copy()
        header_out["CTYPE1"] = "RA---TAN"
        header_out["CTYPE2"] = "DEC--TAN"
        header_out["CRVAL1"] = 266.39311
        header_out["CRVAL2"] = -28.939779

        data = hdu_in.data

    image_stack = np.stack((data, data.T, data[::-1], data[:, ::-1]))

    # Ensure we exercise the bad-value handling modes
    image_stack[1, 30, 30] = np.nan
    image_stack[2, 20, 40] = np.inf

    # Build the reference array through un-broadcast reprojections
    array_ref = np.empty_like(image_stack)
    footprint_ref = np.empty_like(image_stack)
    for i in range(len(image_stack)):
        array_out, footprint_out = reproject_adaptive(
            (image_stack[i], header_in),
            header_out,
            conserve_flux=conserve_flux,
            boundary_mode=boundary_mode,
            kernel=kernel,
            center_jacobian=center_jacobian,
            bad_value_mode=bad_value_mode,
        )
        array_ref[i] = array_out
        footprint_ref[i] = footprint_out

    return image_stack, array_ref, footprint_ref, header_in, header_out


@pytest.mark.parametrize("input_extra_dims", (1, 2))
@pytest.mark.parametrize("output_shape", (None, "single", "full"))
@pytest.mark.parametrize("input_as_wcs", (True, False))
@pytest.mark.parametrize("output_as_wcs", (True, False))
def test_broadcast_reprojection(input_extra_dims, output_shape, input_as_wcs, output_as_wcs):
    image_stack, array_ref, footprint_ref, header_in, header_out = _setup_for_broadcast_test()

    # Test both single and multiple dimensions being broadcast
    if input_extra_dims == 2:
        image_stack = image_stack.reshape((2, 2, *image_stack.shape[-2:]))
        array_ref = array_ref.reshape(image_stack.shape)
        footprint_ref = footprint_ref.reshape(image_stack.shape)

    # Test different ways of providing the output shape
    if output_shape == "single":
        # Have the broadcast dimensions be auto-added to the output shape
        output_shape = image_stack.shape[-2:]
    elif output_shape == "full":
        # Provide the broadcast dimensions as part of the output shape
        output_shape = image_stack.shape

    # Ensure logic works with WCS inputs as well as Header inputs
    if input_as_wcs:
        header_in = WCS(header_in)
    if output_as_wcs:
        header_out = WCS(header_out)
        if output_shape is None:
            # This combination of parameter values is not valid
            return

    array_broadcast, footprint_broadcast = reproject_adaptive(
        (image_stack, header_in),
        header_out,
        output_shape,
        conserve_flux=False,
        boundary_mode="strict",
    )

    np.testing.assert_array_equal(footprint_broadcast, footprint_ref)
    np.testing.assert_allclose(array_broadcast, array_ref)


def _test_broadcast_reprojection_algo_specific_options(
    conserve_flux=False,
    boundary_mode="strict",
    kernel="gaussian",
    center_jacobian=False,
    bad_value_mode="strict",
):
    image_stack, array_ref, footprint_ref, header_in, header_out = _setup_for_broadcast_test(
        conserve_flux=conserve_flux,
        boundary_mode=boundary_mode,
        kernel=kernel,
        center_jacobian=center_jacobian,
        bad_value_mode=bad_value_mode,
    )

    array_broadcast, footprint_broadcast = reproject_adaptive(
        (image_stack, header_in),
        header_out,
        image_stack.shape,
        conserve_flux=conserve_flux,
        boundary_mode=boundary_mode,
        kernel=kernel,
        center_jacobian=center_jacobian,
        bad_value_mode=bad_value_mode,
    )

    np.testing.assert_array_equal(footprint_broadcast, footprint_ref)
    np.testing.assert_allclose(array_broadcast, array_ref)


@pytest.mark.parametrize("kernel", ("gaussian", "hann"))
def test_broadcast_reprojection_kernel(kernel):
    _test_broadcast_reprojection_algo_specific_options(kernel=kernel)


@pytest.mark.parametrize(
    "boundary_mode",
    ("strict", "constant", "grid-constant", "ignore", "ignore_threshold", "nearest"),
)
def test_broadcast_reprojection_boundary_mode(boundary_mode):
    _test_broadcast_reprojection_algo_specific_options(boundary_mode=boundary_mode)


@pytest.mark.parametrize("conserve_flux", (True, False))
def test_broadcast_reprojection_conserve_flux(conserve_flux):
    _test_broadcast_reprojection_algo_specific_options(conserve_flux=conserve_flux)


@pytest.mark.parametrize("center_jacobian", (True, False))
def test_broadcast_reprojection_center_jacobian(center_jacobian):
    _test_broadcast_reprojection_algo_specific_options(center_jacobian=center_jacobian)


@pytest.mark.parametrize("bad_value_mode", ("strict", "ignore", "constant"))
def test_broadcast_reprojection_bad_value_mode(bad_value_mode):
    _test_broadcast_reprojection_algo_specific_options(bad_value_mode=bad_value_mode)


def test_adaptive_input_output_types(
    valid_celestial_input_data, valid_celestial_output_projections
):
    # Check that all valid input/output types work properly

    array_ref, wcs_in_ref, input_value, kwargs_in = valid_celestial_input_data

    wcs_out_ref, shape_ref, output_value, kwargs_out = valid_celestial_output_projections

    # Compute reference

    output_ref, footprint_ref = reproject_adaptive(
        (array_ref, wcs_in_ref), wcs_out_ref, shape_out=shape_ref
    )

    # Compute test

    output_test, footprint_test = reproject_adaptive(
        input_value, output_value, **kwargs_in, **kwargs_out
    )

    assert_allclose(output_ref, output_test)
    assert_allclose(footprint_ref, footprint_test)


def test_readonly_array():
    wcs1 = WCS(naxis=2)
    wcs2 = WCS(naxis=2)
    wcs2.wcs.crpix = [2, 2]

    array = np.random.random((128, 128))
    array.flags.writeable = False

    # This should work
    reproject_adaptive((array, wcs1), wcs2, shape_out=(128, 128))



import numpy as np
from astropy.io import fits

__all__ = ["fits_writeto_withtrim", "fits_getdata_untrimmed"]


def fits_writeto_withtrim(filename, array, header, **kwargs):

    if array.ndim == 3:

        mask = ~np.isnan(array)

        slices = []
        for axis in range(3):
            mask_1d = np.any(mask, axis=tuple(a for a in range(3) if a != axis))
            indices = np.nonzero(mask_1d)[0]
            if len(indices) == 0:
                return
            imin, imax = indices[0], indices[-1]
            slices.append(slice(imin, imax + 1))
            fits_axis = 3 - axis
            # Store the original (untrimmed) size so that a reader can restore
            # the full tile, and record the size of the leading margin removed.
            header[f"ONAXIS{fits_axis}"] = array.shape[axis]
            header[f"TRIM{fits_axis}"] = imin
            header[f"CRPIX{fits_axis}"] -= imin
            header[f"NAXIS{fits_axis}"] = imax + 1 - imin
        array = array[tuple(slices)]

    array = array.astype(np.float32)
    header["BITPIX"] = -32

    fits.writeto(filename, array, header, **kwargs)


def fits_getdata_untrimmed(filename, *, tile_size, tile_depth):

    with fits.open(filename) as hdulist:
        data = hdulist[0].data
        header = hdulist[0].header

    # Prefer the original tile size recorded at write time (ONAXISn), falling
    # back to the requested tile dimensions for tiles written without it.
    onaxis = (
        int(header.get("ONAXIS3", tile_depth)),
        int(header.get("ONAXIS2", tile_size)),
        int(header.get("ONAXIS1", tile_size)),
    )

    pad_before = tuple(header.get(f"TRIM{3 - axis}", 0) for axis in range(3))
    shape = data.shape
    pad_after = (
        onaxis[0] - shape[0] - pad_before[0],
        onaxis[1] - shape[1] - pad_before[1],
        onaxis[2] - shape[2] - pad_before[2],
    )

    data = np.pad(
        data,
        list(zip(pad_before, pad_after, strict=False)),
        mode="constant",
        constant_values=np.nan,
    )

    assert data.shape == onaxis

    return data



import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.ndimage import map_coordinates as scipy_map_coordinates

from reproject._array_utils import dask_map_coordinates, map_coordinates


@pytest.mark.parametrize("cval", [3, np.nan])
@pytest.mark.parametrize("shape", [(3, 4), (30, 40), (3, 40)])
@pytest.mark.parametrize("order", [0, 1, 2, 3])
@pytest.mark.parametrize("dtype", (">f4", ">f8", "<f4", "<f8"))
def test_custom_map_coordinates(cval, shape, order, dtype):
    np.random.seed(1249)

    data = np.random.random(shape).astype(dtype)

    coords = np.random.uniform(-2, max(shape) + 2, (2, 100000))

    expected = scipy_map_coordinates(
        np.pad(data, 1, mode="edge"),
        coords + 1,
        order=order,
        cval=cval,
        mode="constant",
    )

    reset = np.zeros(coords.shape[1], dtype=bool)

    for i in range(coords.shape[0]):
        reset |= coords[i] < -0.5
        reset |= coords[i] > data.shape[i] - 0.5

    expected[reset] = cval

    result1 = map_coordinates(
        data,
        coords,
        order=order,
        cval=cval,
        mode="constant",
    )

    # If order >= 2, the padding we used in the reference result will give
    # subtly different results, so no point in comparing this.
    if order < 2:
        assert_allclose(result1, expected)

    result2 = dask_map_coordinates(
        data,
        coords,
        order=order,
        cval=cval,
        mode="constant",
    )

    assert_allclose(result1, result2)



import numpy as np
import pytest
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS

from reproject import reproject_interp
from reproject.hips import reproject_to_hips
from reproject.hips._dask_array import hips_as_dask_array


class TestHIPSDaskArray:

    def setup_method(self):
        # We use an all-sky WCS image as input since this will test all parts
        # of the HiPS projection (some issues happen around boundaries for instance)
        hdu = fits.open(get_pkg_data_filename("allsky/allsky_rosat.fits"))[0]
        self.original_wcs = WCS(hdu.header)
        self.original_array = hdu.data.size + np.arange(hdu.data.size).reshape(hdu.data.shape)

        self.original_array_3d = self.original_array.reshape((1,) + self.original_array.shape)
        self.original_array_3d = self.original_array_3d * np.arange(1, 11).reshape((10, 1, 1))
        assert self.original_array_3d.shape == (10, 240, 480)

        self.original_wcs_3d = self.original_wcs.sub([1, 2, 0])
        self.original_wcs_3d.wcs.ctype[2] = "FREQ"
        self.original_wcs_3d.wcs.crval[2] = 1e10
        self.original_wcs_3d.wcs.cdelt[2] = 1e9
        self.original_wcs_3d.wcs.crpix[2] = 1
        self.original_wcs_3d._naxis = list(self.original_array_3d.shape[::-1])

    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    @pytest.mark.parametrize("frame", ("galactic", "equatorial"))
    @pytest.mark.parametrize("level", (0, 1))
    @pytest.mark.remote_data
    def test_roundtrip(self, tmp_path, frame, level):

        output_directory = tmp_path / "roundtrip"

        # Note that we always use level=1 to generate, but use a variable level
        # to construct the dask array - this is deliberate and ensure that the
        # dask array has a proper separation of maximum and current level.
        reproject_to_hips(
            (self.original_array, self.original_wcs),
            coord_system_out=frame,
            level=1,
            level_depth=6,
            reproject_function=reproject_interp,
            output_directory=output_directory,
            tile_size=256,
        )

        # Represent the HiPS as a dask array
        dask_array, wcs = hips_as_dask_array(output_directory, level=level)

        # Reproject back to the original WCS
        final_array, footprint = reproject_interp(
            (dask_array, wcs),
            self.original_wcs,
            shape_out=self.original_array.shape,
        )

        # FIXME: Due to boundary effects and the fact there are NaN values in
        # the whole-map dask array, there are a few NaN pixels in the image in
        # the end. For now, we tolerate a small fraction of NaN pixels, and to
        # fix this we should modify the dask array so that in empty tiles
        # adjacent to non-empty tiles, we set the values to the boundaries of
        # the non-empty neighbouring tiles so that interpolation doesn't run
        # into any issues. In theory there should be around 90500 pixels inside
        # the valid region of the image, so we require at least 90000 valid
        # values.

        valid = ~np.isnan(final_array)
        assert np.sum(valid) > 90000
        np.testing.assert_allclose(final_array[valid], self.original_array[valid], rtol=0.01)

    @pytest.mark.remote_data
    def test_level_validation(self, tmp_path):

        output_directory = tmp_path / "levels"

        reproject_to_hips(
            (self.original_array, self.original_wcs),
            coord_system_out="equatorial",
            level=1,
            reproject_function=reproject_interp,
            output_directory=output_directory,
            tile_size=32,
        )

        dask_array, wcs = hips_as_dask_array(output_directory, level=0)
        assert dask_array.shape == (160, 160)

        dask_array, wcs = hips_as_dask_array(output_directory, level=1)
        assert dask_array.shape == (320, 320)

        dask_array, wcs = hips_as_dask_array(output_directory)
        assert dask_array.shape == (320, 320)

        with pytest.raises(Exception, match=r"does not contain spatial level 2 data"):
            hips_as_dask_array(output_directory, level=2)

        with pytest.raises(Exception, match=r"should be positive"):
            hips_as_dask_array(output_directory, level=-1)

    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    @pytest.mark.parametrize("frame", ("galactic", "equatorial"))
    @pytest.mark.parametrize("level", (0, 1))
    @pytest.mark.remote_data
    def test_roundtrip_3d(self, tmp_path, frame, level):

        output_directory = tmp_path / "roundtrip"

        # Note that we always use level=1 to generate, but use a variable level
        # to construct the dask array - this is deliberate and ensure that the
        # dask array has a proper separation of maximum and current level.
        reproject_to_hips(
            (self.original_array_3d, self.original_wcs_3d),
            coord_system_out=frame,
            level=1,
            reproject_function=reproject_interp,
            output_directory=output_directory,
            tile_size=32,
            tile_depth=8,
        )

        # Represent the HiPS as a dask array
        dask_array, wcs = hips_as_dask_array(output_directory, level=level)

        # FIXME: at this point we should be able to do:
        #
        # Reproject back to the original WCS
        # final_array, footprint = reproject_interp(
        #     (dask_array, wcs),
        #     self.original_wcs_3d,
        #     shape_out=self.original_array_3d.shape,
        # )
        #
        # However this does not work properly due to this issue:
        # https://github.com/astropy/astropy/issues/18690
        #
        # For now, we pick a sub-region of the array to check

        subset = (slice(None), slice(50, None), slice(50, None))

        final_array, footprint = reproject_interp(
            (dask_array, wcs),
            self.original_wcs_3d[subset],
            shape_out=self.original_array_3d[subset].shape,
        )

        # NOTE: The two last channels are empty - this is normal and is because
        # of the interpolation on the spectral grid

        valid = ~np.isnan(final_array)[:8]
        assert np.sum(valid) > 450000  # similar to 2D test
        np.testing.assert_allclose(
            final_array[:8][valid],
            self.original_array_3d[subset][:8][valid],
            rtol=0.1 if level == 1 else 0.4,
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from math import exp

import numpy as np

__all__ = ["solve_corrections_sgd", "determine_offset_matrix"]


def determine_offset_matrix(arrays):
    """
    Given a list of ReprojectedArraySubset, determine the offset
    matrix between all arrays.

    Parameters
    ----------
    arrays : list
        The list of ReprojectedArraySubset objects to determine the offsets for.

    Returns
    -------
    `numpy.ndarray`
        The offset matrix.
    """

    N = len(arrays)

    # Set up matrix to record differences
    offset_matrix = np.ones((N, N)) * np.nan

    # Loop over all pairs of images and check for overlap
    for i1, array1 in enumerate(arrays):
        for i2, array2 in enumerate(arrays):
            if i2 <= i1:
                continue
            if array1.overlaps(array2):
                difference = array1 - array2
                if np.any(difference.footprint):
                    values = difference.array[difference.footprint]
                    offset_matrix[i1, i2] = np.median(values)
                    offset_matrix[i2, i1] = -offset_matrix[i1, i2]

    return offset_matrix


def solve_corrections_sgd(offset_matrix, eta_initial=1, eta_half_life=100, rtol=1e-10, atol=0):
    r"""
    Given a matrix of offsets from each image to each other image, find the
    optimal offsets to use using Stochastic Gradient Descent.

    Given N images, we can construct an NxN matrix Oij giving the typical (e.g.
    mean, median, or other statistic) offset from each image to each other
    image. This can be a reasonably sparse matrix since not all images
    necessarily overlap. From this we then want to find a vector of N
    corrections Ci to apply to each image to minimize the differences.

    We do this by using the Stochastic Gradient Descent algorithm:

    https://en.wikipedia.org/wiki/Stochastic_gradient_descent

    Essentially what we are trying to minimize is the difference between Dij
    and a matrix of the same shape constructed from the Oi values.

    The learning rate is decreased using a decaying exponential:

        $$\eta = \eta_{\rm initial} * \exp{(-i/t_{\eta})}$$

    Parameters
    ----------
    offset_matrix : `~numpy.ndarray`
        The NxN matrix giving the offsets between all images (or NaN if
        an offset could not be determined).
    eta_initial : float
        The initial learning rate to use.
    eta_half_life : float
        The number of iterations after which the learning rate should be
        decreased by a factor $e$.
    rtol : float
        The relative tolerance to use to determine if the corrections have
        converged.
    atol : float
        The absolute tolerance to use to determine if the corrections have
        converged.

    Returns
    -------
    `numpy.ndarray`
        The corrections for each frame.
    """

    if offset_matrix.ndim != 2 or offset_matrix.shape[0] != offset_matrix.shape[1]:
        raise ValueError("offset_matrix should be a square NxN matrix")

    N = offset_matrix.shape[0]

    indices = np.arange(N)
    corrections = np.zeros(N)

    # Keep track of previous corrections to know whether the algorithm
    # has converged
    previous_corrections = None

    for iteration in range(int(eta_half_life * 10)):
        # Shuffle the indices to avoid cyclical behavior
        np.random.shuffle(indices)

        # Update learning rate
        eta = eta_initial * exp(-iteration / eta_half_life)

        # Loop over each index and update the offset. What we call rows and
        # columns is arbitrary, but for the purpose of the comments below, we
        # treat this as iterating over rows of the matrix.
        for i in indices:
            if np.isnan(corrections[i]):
                continue

            # Since the matrix might be sparse, we consider only columns which
            # are not NaN
            keep = ~np.isnan(offset_matrix[i, :])

            # Compute the row of the offset matrix one would get with the
            # current offsets
            fitted_offset_matrix_row = corrections[i] - corrections[keep]

            # The difference between the actual row in the matrix and this
            # fitted row gives us a measure of the gradient, so we then
            # adjust the solution in that direction.
            corrections[i] += eta * np.mean(offset_matrix[i, keep] - fitted_offset_matrix_row)

        # Subtract the mean offset from the offsets to make sure that the
        # corrections stay centered around zero
        corrections -= np.nanmean(corrections)

        if previous_corrections is not None:
            if np.allclose(corrections, previous_corrections, rtol=rtol, atol=atol):
                break  # the algorithm has converged

        previous_corrections = corrections.copy()

    return corrections



import os
import shutil
import uuid
import warnings
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from datetime import UTC, datetime
from itertools import product
from logging import getLogger
from pathlib import Path

import numpy as np
from astropy import units as u
from astropy.coordinates import (
    ICRS,
    BarycentricTrueEcliptic,
    Galactic,
)
from astropy.io import fits
from astropy.nddata import block_reduce
from astropy_healpix import (
    HEALPix,
    level_to_nside,
    nside_to_level,
    pixel_resolution_to_nside,
)
from PIL import Image

from .._array_utils import sample_array_edges
from .._wcs_utils import has_celestial, has_spectral, pixel_scale
from ..utils import as_transparent_rgb, is_jpeg, is_png, parse_input_data
from ._trim_utils import fits_getdata_untrimmed, fits_writeto_withtrim
from ._utils import (
    load_properties,
    make_tile_folders,
    save_properties,
    skycoord_first,
    spectral_coord_to_index,
    tile_filename,
    tile_header,
)

__all__ = ["reproject_from_hips", "reproject_to_hips", "coadd_hips"]


INDEX_HTML = """
<!DOCTYPE html>

<html>
<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, height=device-height, initial-scale=1.0, user-scalable=no">


    <script src="https://aladin.cds.unistra.fr/hips-templates/hips-landing-page.js" type="text/javascript"></script>
    <noscript>Please enable Javascript to view this page.</noscript>
</head>

<body></body>

<script type="text/javascript">
    buildLandingPage({alScriptURL: 'https://aladin.cds.unistra.fr/AladinLite/api/v3/3.7.2-beta/aladin.js'});
</script>

</html>
"""

RESERVED_PROPERTIES = [
    "dataproduct_type",
    "hips_version",
    "hips_tile_format",
    "hips_tile_width",
    "hips_order",
    "hips_frame",
]

VALID_COORD_SYSTEM = {
    "equatorial": ICRS(),
    "galactic": Galactic(),
    "ecliptic": BarycentricTrueEcliptic(),
}

VALID_TILE_FORMATS = {"fits", "png", "jpeg"}


EXTENSION = {"fits": "fits", "png": "png", "jpeg": "jpg"}


def reproject_from_hips():
    raise NotImplementedError()


def reproject_to_hips(
    input_data,
    *,
    coord_system_out,
    reproject_function,
    output_directory,
    level=None,
    level_depth=None,
    hdu_in=0,
    tile_size=512,
    tile_depth=16,
    progress_bar=None,
    threads=False,
    properties=None,
    **kwargs,
):
    """
    Reproject data from a standard projection to a set of Hierarchical Progressive
    Surveys (HiPS) tiles.

    Parameters
    ----------
    input_data : object
        The input data to reproject. This can be:

            * The name of a FITS file as a `str` or a `pathlib.Path` object
            * An `~astropy.io.fits.HDUList` object
            * An image HDU object such as a `~astropy.io.fits.PrimaryHDU`,
              `~astropy.io.fits.ImageHDU`, or `~astropy.io.fits.CompImageHDU`
              instance
            * A tuple where the first element is a `~numpy.ndarray` and the
              second element is either a
              `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
              `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a
              `~astropy.io.fits.Header` object
            * An `~astropy.nddata.NDData` object from which the ``.data`` and
              ``.wcs`` attributes will be used as the input data.
            * The name of a PNG or JPEG file with AVM metadata

    coord_system_out : {'equatorial', 'galactic', 'ecliptic' }
        The target coordinate system for the HEALPIX projection
    reproject_function : callable
        The function to use for the reprojection.
    output_directory : str
        The name of the output directory - if this already exists, an error
        will be raised.
    level : int, optional
        The number of levels of tiles for celestial coordinates.
    level_depth : int, optional
        The number of levels of tiles for the third (e.g. spectral) dimension.
    hdu_in : int or str, optional
        If ``input_data`` is a FITS file or an `~astropy.io.fits.HDUList`
        instance, specifies the HDU to use.
    tile_size : int or tuple, optional
        The size of each individual tile (defaults to 512) for celestial dimensions.
    tile_depth : int or tuple, optional
        The depth of each individual tile (defaults to 16) for the third (e.g. spectral) dimension when present.
    progress_bar : callable, optional
        If specified, use this as a progress_bar to track loop iterations over
        data sets.
    threads : bool or int
        If `False`, no multi-threading is used. If an integer, this number of
        threads will be used, and if `True`, the number of threads will be chosen
        automatically.
    properties : dict, optional
        Dictionary of properties that should be output to the ``properties``
        file inside the HiPS dataset. At list of properties and their meanings
        can be found in the `HiPS 1.0 <https://www.ivoa.net/documents/HiPS/20170406/PR-HIPS-1.0-20170406.pdf>`_
        description.
    **kwargs
        Keyword arguments to be passed to the reprojection function.

    Returns
    -------
    None
        This function does not return a value.
    """

    tile_format = "fits"

    if isinstance(input_data, str | Path):
        if is_png(input_data):
            tile_format = "png"
        elif is_jpeg(input_data):
            tile_format = "jpeg"

    array_in, wcs_in = parse_input_data(input_data, hdu_in=hdu_in)

    if not has_celestial(wcs_in):
        raise Exception("Only data with a celestial WCS can be reprojected to HiPS tiles")

    if wcs_in.low_level_wcs.pixel_n_dim != wcs_in.low_level_wcs.world_n_dim:
        raise Exception(
            f"Number of pixel ({wcs_in.low_level_wcs.pixel_n_dim}) "
            f"and world ({wcs_in.low_level_wcs.world_n_dim}) "
            f"dimensions do not match"
        )

    if wcs_in.low_level_wcs.pixel_n_dim == 2:
        ndim = 2
    elif wcs_in.low_level_wcs.pixel_n_dim == 3:
        if has_spectral(wcs_in):
            ndim = 3
        else:
            raise NotImplementedError(
                "Only 3-d data with a spectral axis are supported at this time"
            )
    else:
        raise Exception("Can only reproject data with 2-d or 3-d WCS")

    if ndim == 2:
        if array_in.ndim not in (2, 3):
            raise Exception("Input array should have 2 or 3 dimensions for 2-dimensional input WCS")
    else:
        if array_in.ndim != ndim:
            raise Exception(
                f"Input array dimensionality ({array_in.ndim}) should match WCS dimensionality ({ndim})"
            )

    if ndim == 3 and tile_format != "fits":
        raise ValueError("Only FITS tiles are supported in HiPS3D mode")

    logger = getLogger(__name__)

    # Check tile size is even
    if tile_size % 2 != 0:
        raise ValueError("tile_size should be even")

    # Check coordinate system
    if coord_system_out in VALID_COORD_SYSTEM:
        frame = VALID_COORD_SYSTEM[coord_system_out]
    else:
        raise ValueError("coord_system_out should be one of " + "/".join(VALID_COORD_SYSTEM))

    # Check tile format
    if tile_format not in VALID_TILE_FORMATS:
        raise ValueError("tile_format should be one of " + "/".join(VALID_TILE_FORMATS))

    if properties is None:
        properties = {}

    if progress_bar is None:
        progress_bar = lambda x: x

    # Determine celestial level if not specified

    if level is None:
        scale = pixel_scale(wcs_in, array_in.shape)
        nside = pixel_resolution_to_nside(scale * tile_size)
        level = int(nside_to_level(nside))
        logger.info(f"Automatically set the HEALPIX level to {level}")

    # Create output directory (and error if it already exists)
    os.makedirs(output_directory, exist_ok=False)

    # Determine center of image and radius to furthest corner, to determine
    # which HiPS tiles need to be generated

    # TODO: this will fail for e.g. allsky maps

    ny, nx = array_in.shape[-2:]

    centers = [(s - 1) / 2 for s in array_in.shape[-wcs_in.pixel_n_dim :]][::-1]
    edges = sample_array_edges(array_in.shape[-wcs_in.pixel_n_dim :], n_samples=2)[::-1]

    if ndim == 2:
        cen_skycoord = wcs_in.pixel_to_world(*centers)
        cor_skycoord = wcs_in.pixel_to_world(*edges)
    else:
        cen_skycoord, _ = skycoord_first(wcs_in.pixel_to_world(*centers))
        cor_skycoord, cor_spectralcoord = skycoord_first(wcs_in.pixel_to_world(*edges))

    separations = cor_skycoord.separation(cen_skycoord)

    if np.any(np.isnan(separations)):

        # At least one of the corners is outside of the region of validity of
        # the WCS, so we use a different approach where we randomly sample a
        # number of positions in the image and then check the maximum
        # separation between any pair of points.

        n_ran = 1000
        ran_x = np.random.uniform(-0.5, nx - 0.5, n_ran)
        ran_y = np.random.uniform(-0.5, nx - 0.5, n_ran)

        if ndim == 2:
            ran_world = wcs_in.pixel_to_world(ran_x, ran_y)
        elif ndim == 3:
            ran_world, _ = skycoord_first(wcs_in.pixel_to_world(ran_x, ran_y, np.zeros(n_ran)))

        separations = ran_world[:, None].separation(ran_world[None, :])

        max_separation = np.nanmax(separations)

    else:

        max_separation = separations.max()

    radius = 1.5 * max_separation

    # TODO: in future if astropy-healpix implements polygon searches, we could
    # use that instead

    # Determine all the celestial indices at the highest level

    nside = level_to_nside(level)
    hp = HEALPix(nside=nside, order="nested", frame=frame)

    if radius > 120 * u.deg:
        indices = np.arange(hp.npix)
    else:
        indices = hp.cone_search_skycoord(cen_skycoord, radius=radius)

    spatial_level = level

    if ndim == 3:

        # If depth level has not been specified, try and determine it
        if level_depth is None:

            for level_depth in range(52):  # FREQ_MAX_ORDER
                spectral_indices_edges = spectral_coord_to_index(level_depth, cor_spectralcoord)
                if np.ptp(spectral_indices_edges) > array_in.shape[0]:
                    break
            else:
                raise Exception(
                    "Could not determine depth level automatically, specify manually with level_depth="
                )

            level_depth = max(0, level_depth - int(np.log2(tile_depth)))

            logger.info(f"Automatically set the Spectral level to {level_depth}")

        # Determine all the spectral indices at the highest spectral level
        spectral_indices_edges = spectral_coord_to_index(level_depth, cor_spectralcoord)
        spectral_indices = np.arange(spectral_indices_edges.min(), spectral_indices_edges.max() + 1)
        indices = [
            (int(idx), int(spec_idx)) for (idx, spec_idx) in product(indices, spectral_indices)
        ]
        level = (level, level_depth)
        tile_dims = (tile_size, tile_depth)

    else:

        tile_dims = tile_size

    logger.info(f"Found {len(indices)} tiles (at most) to generate at level {level}")

    # PERF: the code above may be prohibitive for large numbers of tiles,
    # so we may want to find a way to iterate over these in chunks.

    # Make all the folders required for the tiles
    make_tile_folders(level=level, indices=indices, output_directory=output_directory)

    # Iterate over the tiles and generate them
    def process(index):
        if hasattr(wcs_in, "deepcopy"):
            wcs_in_copy = wcs_in.deepcopy()
        else:
            wcs_in_copy = deepcopy(wcs_in)

        header = tile_header(level=level, index=index, frame=frame, tile_dims=tile_dims)

        if isinstance(header, tuple):

            array_out1, footprint1 = reproject_function(
                (array_in, wcs_in_copy), header[0], **kwargs
            )
            array_out2, footprint2 = reproject_function(
                (array_in, wcs_in_copy), header[1], **kwargs
            )
            with np.errstate(invalid="ignore"):
                array_out = (
                    np.nan_to_num(array_out1) * footprint1 + np.nan_to_num(array_out2) * footprint2
                ) / (footprint1 + footprint2)
                footprint = (footprint1 + footprint2) / 2
            header = header[0]
        else:
            array_out, footprint = reproject_function((array_in, wcs_in_copy), header, **kwargs)

        if np.all(footprint == 0):
            return None

        if tile_format == "fits":
            array_out[footprint == 0] = np.nan
            pixel_min = np.nanmin(array_out)
            pixel_max = np.nanmax(array_out)
            fits_writeto_withtrim(
                tile_filename(
                    level=level,
                    index=index,
                    output_directory=output_directory,
                    extension=EXTENSION[tile_format],
                ),
                array_out,
                header,
            )

        else:
            if tile_format == "png":
                image = as_transparent_rgb(array_out, alpha=footprint[0])
            else:
                array_out[np.isnan(array_out)] = 0.0
                image = as_transparent_rgb(array_out).convert("RGB")
            image.save(
                tile_filename(
                    level=level,
                    index=index,
                    output_directory=output_directory,
                    extension=EXTENSION[tile_format],
                )
            )
            pixel_min, pixel_max = None, None

        return index, pixel_min, pixel_max

    if tile_format == "fits":
        pixel_min = np.inf
        pixel_max = -np.inf

    if threads:
        generated_indices = []
        with ThreadPoolExecutor(max_workers=None if threads is True else threads) as executor:
            futures = [executor.submit(process, index) for index in indices]
            for future in progress_bar(futures):
                result = future.result()
                if result is not None:
                    generated_indices.append(result[0])
                    if tile_format == "fits":
                        pixel_min = min(pixel_min, result[1])
                        pixel_max = max(pixel_max, result[2])
    else:
        generated_indices = []
        for index in progress_bar(indices):
            result = process(index)
            if result is not None:
                generated_indices.append(result[0])
                if tile_format == "fits":
                    pixel_min = min(pixel_min, result[1])
                    pixel_max = max(pixel_max, result[2])

    indices = generated_indices

    # Generate properties file

    cen_icrs = cen_skycoord.icrs

    for key in properties:
        if key in RESERVED_PROPERTIES:
            raise ValueError(f"Cannot override property {key}")

    generated_properties = {
        "creator_did": f"ivo://reproject/P/{str(uuid.uuid4())}",
        "obs_title": os.path.basename(os.path.normpath(output_directory)),
        "hips_version": "1.4",
        "hips_release_date": datetime.now(UTC).strftime("%Y-%m-%dT%H:%MZ"),
        "hips_status": "public master clonableOnce",
        "hips_tile_format": tile_format,
        "hips_tile_width": tile_size,
        "hips_order": spatial_level,
        "hips_frame": coord_system_out,
        "hips_builder": "astropy/reproject",
        "hips_initial_ra": cen_icrs.ra.deg,
        "hips_initial_dec": cen_icrs.dec.deg,
        "hips_initial_fov": radius.deg,
    }

    if ndim == 2:
        generated_properties["dataproduct_type"] = "image"
    else:
        generated_properties["dataproduct_type"] = "spectral-cube"
        generated_properties["hips_order_axis2"] = level_depth
        generated_properties["hips_order_min"] = 0
        generated_properties["hips_tile_depth"] = tile_depth
        wav = cor_spectralcoord.to_value(u.m)
        generated_properties["em_min"] = wav.min()
        generated_properties["em_max"] = wav.max()

    if tile_format == "fits":
        generated_properties["hips_pixel_bitpix"] = -32
        if not np.isinf(pixel_min) and not np.isinf(pixel_max):
            generated_properties["hips_pixel_cut"] = f"{pixel_min} {pixel_max}"

    generated_properties.update(properties)

    save_properties(output_directory, generated_properties)

    save_index(output_directory)

    compute_lower_resolution_tiles(
        output_directory=output_directory,
        ndim=ndim,
        frame=frame,
        tile_format=tile_format,
        tile_size=tile_size,
        tile_depth=tile_depth,
        spatial_level=spatial_level,
        level_depth=level_depth,
    )


def find_indices(*, output_directory, ndim, spatial_level, level_depth):

    if ndim == 2:

        norder_directory = os.path.join(
            output_directory,
            f"Norder{spatial_level}",
        )

        for _, _, filenames in os.walk(norder_directory):
            for filename in filenames:
                yield int(filename.split(".")[0].replace("Npix", ""))

    else:

        norder_directory = os.path.join(
            output_directory,
            f"Norder{spatial_level}_{level_depth}",
        )

        for _, _, filenames in os.walk(norder_directory):
            for filename in filenames:
                indices = filename.split(".")[0].replace("Npix", "").split("_")
                yield int(indices[0]), int(indices[1])


def compute_lower_resolution_tiles(
    *,
    output_directory,
    ndim,
    frame,
    tile_format,
    tile_size,
    tile_depth,
    spatial_level,
    level_depth,
    indices=None,
):

    if ndim == 3:
        tile_dims = tile_size, tile_depth
    else:
        tile_dims = tile_size

    # Iterate over higher levels and compute lower resolution tiles

    half_tile_size = tile_size // 2
    if ndim == 3:
        half_tile_depth = tile_depth // 2

    if indices is None:
        indices = list(
            find_indices(
                ndim=ndim,
                output_directory=output_directory,
                spatial_level=spatial_level,
                level_depth=level_depth,
            )
        )

    # If no tiles were generated at the deepest level there is nothing to
    # build lower-resolution tiles from.
    if len(indices) == 0:
        warnings.warn(
            "No tiles were generated at the deepest level, so no "
            "lower-resolution tiles will be produced",
            stacklevel=2,
        )
        return

    for sub in range(1, spatial_level + 1):

        if ndim == 2:
            ilevel = spatial_level - sub
        else:
            # The spectral order is tied to the spatial order and decreases by
            # one at each step, but is clamped at 0: once the coarsest spectral
            # order is reached, lower-resolution tiles only degrade spatially.
            prev_spectral_level = max(0, level_depth - (sub - 1))
            spectral_level = max(0, level_depth - sub)
            spectral_changed = spectral_level != prev_spectral_level
            ilevel = (spatial_level - sub, spectral_level)

        # Find index of tiles to produce at lower-resolution levels
        if ndim == 2:
            indices = np.sort(np.unique(np.asarray(indices) // 4))
        elif spectral_changed:
            indices = sorted(
                set(
                    [
                        (spatial_index // 4, spectral_index // 2)
                        for (spatial_index, spectral_index) in indices
                    ]
                )
            )
        else:
            indices = sorted(
                set(
                    [
                        (spatial_index // 4, spectral_index)
                        for (spatial_index, spectral_index) in indices
                    ]
                )
            )

        make_tile_folders(level=ilevel, indices=indices, output_directory=output_directory)

        for index in indices:

            header = tile_header(level=ilevel, index=index, frame=frame, tile_dims=tile_dims)

            if isinstance(header, tuple):
                header = header[0]

            if ndim == 2:

                if tile_format == "fits":
                    array = np.zeros((tile_size, tile_size)) * np.nan
                elif tile_format == "png":
                    array = np.zeros((tile_size, tile_size, 4), dtype=np.uint8)
                else:
                    array = np.zeros((tile_size, tile_size, 3), dtype=np.uint8)

                for subindex in range(4):

                    current_index = 4 * index + subindex
                    subtile_filename = tile_filename(
                        level=ilevel + 1,
                        index=current_index,
                        output_directory=output_directory,
                        extension=EXTENSION[tile_format],
                    )

                    if os.path.exists(subtile_filename):

                        if tile_format == "fits":
                            tile_data = fits.getdata(subtile_filename)
                            # np.nanmean can emit warnings, so ignore these
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore")
                                data = block_reduce(tile_data, 2, func=np.nanmean)
                        else:
                            data = block_reduce(
                                np.array(Image.open(subtile_filename))[::-1],
                                (2, 2, 1),
                                func=np.mean,
                            )

                        if subindex == 0:
                            array[half_tile_size:, :half_tile_size] = data
                        elif subindex == 2:
                            array[half_tile_size:, half_tile_size:] = data
                        elif subindex == 1:
                            array[:half_tile_size, :half_tile_size] = data
                        elif subindex == 3:
                            array[:half_tile_size, half_tile_size:] = data

            elif ndim == 3:

                array = np.ones((tile_depth, tile_size, tile_size)) * np.nan

                # When the spectral order is unchanged (clamped at 0) we only
                # combine the four spatial sub-tiles and reduce spatially.
                spec_subindices = range(2) if spectral_changed else range(1)

                for subindex in range(4):
                    for subindex_spec in spec_subindices:

                        if spectral_changed:
                            current_spectral_index = 2 * index[1] + subindex_spec
                        else:
                            current_spectral_index = index[1]

                        current_index = (4 * index[0] + subindex, current_spectral_index)
                        subtile_filename = tile_filename(
                            level=(ilevel[0] + 1, prev_spectral_level),
                            index=current_index,
                            output_directory=output_directory,
                            extension=EXTENSION[tile_format],
                        )

                        if os.path.exists(subtile_filename):

                            # np.nanmean can emit warnings, so ignore these
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore")
                                data = block_reduce(
                                    fits_getdata_untrimmed(
                                        subtile_filename, tile_size=tile_size, tile_depth=tile_depth
                                    ),
                                    2 if spectral_changed else (1, 2, 2),
                                    func=np.nanmean,
                                )

                            if not spectral_changed:
                                subtile_slice = [slice(None)]
                            elif subindex_spec == 0:
                                subtile_slice = [slice(None, half_tile_depth)]
                            else:
                                subtile_slice = [slice(half_tile_depth, None)]

                            if subindex == 0:
                                subtile_slice.extend(
                                    [slice(half_tile_size, None), slice(None, half_tile_size)]
                                )
                            elif subindex == 2:
                                subtile_slice.extend(
                                    [slice(half_tile_size, None), slice(half_tile_size, None)]
                                )
                            elif subindex == 1:
                                subtile_slice.extend(
                                    [slice(None, half_tile_size), slice(None, half_tile_size)]
                                )
                            elif subindex == 3:
                                subtile_slice.extend(
                                    [slice(None, half_tile_size), slice(half_tile_size, None)]
                                )

                            array[tuple(subtile_slice)] = data

            if tile_format == "fits":
                fits_writeto_withtrim(
                    tile_filename(
                        level=ilevel,
                        index=index,
                        output_directory=output_directory,
                        extension=EXTENSION[tile_format],
                    ),
                    array,
                    header,
                )
            else:
                image = as_transparent_rgb(array.transpose(2, 0, 1))
                if tile_format == "jpeg":
                    image = image.convert("RGB")
                image.save(
                    tile_filename(
                        level=ilevel,
                        index=index,
                        output_directory=output_directory,
                        extension=EXTENSION[tile_format],
                    )
                )


def save_index(directory):
    with open(os.path.join(directory, "index.html"), "w") as f:
        f.write(INDEX_HTML)


def coadd_hips(input_directories, output_directory):
    """
    Given multiple HiPS directories, combine these into a single HiPS.

    The coordinate frame and tile format of the different input directories
    should match.

    In cases of overlap, the last image in the order of input_directories is used.

    Parameters
    ----------
    input_directories : iterable
        Iterable of HiPS directory names.
    output_directory : str
        The path to the output directory.
    """

    all_properties = [load_properties(directory) for directory in input_directories]

    tile_formats = [p["hips_tile_format"] for p in all_properties]
    hips_frame = [p["hips_frame"] for p in all_properties]
    hips_order = [p["hips_order"] for p in all_properties]

    if len(set(tile_formats)) > 1:
        raise ValueError(f"tile_format values do not match: {tile_formats}")
    else:
        tile_format = tile_formats[0]

    if len(set(hips_frame)) > 1:
        raise ValueError(f"tile_format values do not match: {hips_frame}")

    reference_properties = all_properties[0]
    reference_properties["hips_order"] = max(hips_order)

    # Create output directory (and error if it already exists)
    os.makedirs(output_directory, exist_ok=False)

    for directory in input_directories:

        for dirpath, _, filenames in os.walk(directory):
            for filename in filenames:
                if not filename.endswith("." + tile_format):
                    continue
                filepath = os.path.join(dirpath, filename)
                target_directory = os.path.join(
                    output_directory, os.path.relpath(dirpath, directory)
                )
                target_filepath = os.path.join(target_directory, filename)
                if not os.path.exists(target_directory):
                    os.makedirs(target_directory)
                if os.path.exists(target_filepath):
                    if tile_format == "png":
                        image1 = Image.open(filepath).convert("RGBA")
                        image2 = Image.open(target_filepath).convert("RGBA")
                        result = Image.alpha_composite(image1, image2)
                        result.save(target_filepath)
                    elif tile_format == "jpeg":
                        raise NotImplementedError(
                            "Convert jpg to png to allow for blending/coadding"
                        )
                    else:
                        raise NotImplementedError()
                else:
                    shutil.copyfile(filepath, target_filepath)

    save_properties(output_directory, reference_properties)

    save_index(output_directory)
